# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for transforming metadata."""


from metldata.builtin_transformations.infer_references.path.resolve import (
    resolve_reference_for_metadata_resource,
)
from metldata.builtin_transformations.infer_references.reference import (
    InferredReference,
)
from metldata.metadata_utils import (
    SelfIdLookUpError,
    get_resources_of_class,
    lookup_self_id,
    upsert_resources_in_metadata,
)
from metldata.model_utils.anchors import (
    AnchorPoint,
    AnchorPointNotFoundError,
    lookup_anchor_point,
)
from metldata.transform.base import (
    Json,
    MetadataModelTransformationError,
    MetadataTransformationError,
)


def add_reference_to_metadata_resource(
    resource: Json,
    global_metadata: Json,
    reference: InferredReference,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Add an inferred reference to an individual metadata resource.

    Args:
        resource: The metadata resource to modify.
        global_metadata: The global metadata context to look up references in.
        reference: The inferred reference.
        anchor_points: The anchor points of the metadata model.

    Raises:
        MetadataTransformationError:
            if the transformation of the metadata fails.
    """

    try:
        target_anchor_point = lookup_anchor_point(
            class_name=reference.target, anchor_points_by_target=anchor_points_by_target
        )
    except AnchorPointNotFoundError as error:
        raise MetadataModelTransformationError(
            f"Cannot add reference '{reference}' to metadata resource '{resource}'"
            + " because the target anchor point could not be found."
        ) from error

    if reference.new_slot in resource:
        raise MetadataModelTransformationError(
            f"Cannot add reference '{reference}' to metadata resource '{resource}'"
            + f" because the target slot '{reference.new_slot}' already exists."
        )

    target_resources = resolve_reference_for_metadata_resource(
        resource=resource,
        global_metadata=global_metadata,
        reference_path=reference.path,
        anchor_points_by_target=anchor_points_by_target,
    )

    # get IDs of final target resources:
    target_ids: set[str] = set()
    for target_resource in target_resources:
        try:
            target_ids.add(
                lookup_self_id(
                    resource=target_resource,
                    identifier_slot=target_anchor_point.identifier_slot,
                )
            )
        except SelfIdLookUpError as error:
            raise MetadataTransformationError(
                f"Cannot add reference '{reference}' to metadata resource '{resource}'"
                + f" because the target resource '{target_resource}' does not have"
                + f" an identifier in slot '{target_anchor_point.identifier_slot}'."
            ) from error

    # add the target IDs to the source resource:
    resource_copy = resource.copy()
    resource_copy[reference.new_slot] = sorted(target_ids)

    return resource_copy


def add_reference_to_metadata(
    *,
    metadata: Json,
    reference: InferredReference,
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Transform metadata by adding an inferred reference.

    Raises:
            MetadataTransformationError:
                if the transformation of the metadata fails.
    """

    resources = get_resources_of_class(
        global_metadata=metadata,
        class_name=reference.source,
        anchor_points_by_target=anchor_points_by_target,
    )

    modified_resources = [
        add_reference_to_metadata_resource(
            resource=resource,
            global_metadata=metadata,
            reference=reference,
            anchor_points_by_target=anchor_points_by_target,
        )
        for resource in resources
    ]

    return upsert_resources_in_metadata(
        resources=modified_resources,
        class_name=reference.source,
        global_metadata=metadata,
        anchor_points_by_target=anchor_points_by_target,
    )


def add_references_to_metadata(
    *,
    metadata: Json,
    references: list[InferredReference],
    anchor_points_by_target: dict[str, AnchorPoint],
) -> Json:
    """Transform metadata and return the transformed one.

    Raises:
        MetadataTransformationError:
            if the transformation of the metadata fails.
    """

    for reference in references:
        metadata = add_reference_to_metadata(
            metadata=metadata,
            reference=reference,
            anchor_points_by_target=anchor_points_by_target,
        )

    return metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A transformation that normalizes a model to a canonical form."""

from metldata.builtin_transformations.normalize_model.config import NormalizationConfig
from metldata.builtin_transformations.normalize_model.model_transform import (
    normalize_model,
)
from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import MetadataTransformer, TransformationDefinition


# pylint: disable=unused-argument
def check_model_assumptions(model: MetadataModel, config: NormalizationConfig):
    """Check that the classes and slots specified in the config exist in the model."""


# pylint: disable=unused-argument
def transform_model(model: MetadataModel, config: NormalizationConfig) -> MetadataModel:
    """Normalize the model."""

    return normalize_model(model)


class NormalizationTransformer(MetadataTransformer[NormalizationConfig]):
    """Transformer for normalizing the metadata model."""

    def transform(self, *, metadata: Json, annotation: SubmissionAnnotation) -> Json:
        """Transforms metadata.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """

        return metadata


NORMALIZATION_TRANSFORMATION = TransformationDefinition[NormalizationConfig](
    config_cls=NormalizationConfig,
    check_model_assumptions=check_model_assumptions,
    transform_model=transform_model,
    metadata_transformer_factory=NormalizationTransformer,
)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing the accession handler."""

from metldata.accession_registry.accession_store import AccessionStore
from metldata.config import SubmissionConfig
from tests.fixtures.config import config_sub_fixture  # noqa: F401


def test_accession_store_happy(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
):
    """Test accession store happy path."""

    expected_accessions = ["accession001", "accession002"]

    accession_store = AccessionStore(config=config_sub_fixture)

    # save accessions
    for expected_accession in expected_accessions:
        accession_store.save(accession=expected_accession)

    # check that the expected accessions exist:
    for expected_accession in expected_accessions:
        assert accession_store.exists(accession=expected_accession)


def test_accession_store_unkown_accession(
    config_sub_fixture: SubmissionConfig,  # noqa: F811
):
    """Test accession store happy path."""

    unknown_accession = "accession003"

    accession_store = AccessionStore(config=config_sub_fixture)
    assert not accession_store.exists(accession=unknown_accession)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Model transformation for aggregate transformations."""

import itertools
from collections import defaultdict
from dataclasses import asdict
from typing import Iterable, Optional

from linkml_runtime.linkml_model import ClassDefinition, SlotDefinition
from stringcase import snakecase

from metldata.builtin_transformations.aggregate.config import (
    AggregateConfig,
    Aggregation,
)
from metldata.builtin_transformations.aggregate.models import (
    MinimalClass,
    MinimalLinkMLModel,
    MinimalNamedSlot,
    MinimalSlot,
)
from metldata.model_utils.anchors import AnchorPoint
from metldata.model_utils.essentials import ROOT_CLASS, MetadataModel
from metldata.model_utils.identifiers import get_class_identifier
from metldata.model_utils.manipulate import add_anchor_point, upsert_class_slot
from metldata.transform.base import MetadataModelTransformationError

Path = tuple[Optional[str], ...]


class PathMatrix:
    """Represents a matrix of data paths."""

    __paths: list[Path]
    leaf_slots: list[MinimalNamedSlot]

    def path_leaves(self) -> Iterable[str]:
        """Yields the path leaves of all paths"""
        for path in self.__paths:
            for node in reversed(path):
                if node is not None:
                    yield node
                    break

    def validate_paths(self) -> None:
        """Validates that the specified output paths are compatible, i.e. that
        no path is a prefix of another path. Note that this implicitly includes
        checking for identical paths.

        Raises:
            MetadataModelTransformationError: If one path is a prefix of another."""
        for path_a, path_b in itertools.combinations(self.__paths, r=2):
            min_len = min(len(path_a), len(path_b))
            if path_a[:min_len] == path_b[:min_len]:
                raise MetadataModelTransformationError(
                    "Incompatible output paths:"
                    f" '{'.'.join(str(elem) for elem in path_a)}"
                    f" and {'.'.join(str(elem) for elem in path_b)}."
                )

    def load_leaf_slots(self, leaf_slots: list[MinimalSlot]) -> None:
        """Combines the paths and the specified leaf slots to NamedSlots and
        stores them."""
        self.leaf_slots = [
            MinimalNamedSlot(slot_name=slot_name, **asdict(slot))
            for slot_name, slot in zip(self.path_leaves(), leaf_slots)
        ]

    @property
    def paths(self):
        """Returns a copy of the paths."""
        return self.__paths[:]

    @property
    def max_depth(self):
        """The maximum depth among all paths"""
        # All path lengths are equal after normalization
        return len(self.__paths[0]) if self.__paths else 0

    def normalize_path_matrix(self) -> None:
        """Equalize the lengths of all paths by appending None values to paths
        shorter than then maximum path length."""
        max_depth = max(len(path) for path in self.__paths)
        self.__paths = [
            path + (None,) * (max_depth - len(path)) for path in self.__paths
        ]

    def add_path(self, path: Path, leaf_slot: MinimalNamedSlot) -> None:
        """Add a path"""
        for existing_path in self.__paths:
            if path == existing_path[: len(path)]:
                raise MetadataModelTransformationError(
                    f"Cannot add conflicting path {path}."
                )
        self.__paths.append(path)
        self.leaf_slots.append(leaf_slot)
        self.normalize_path_matrix()

    def load_path_strings(self, paths: list[list[str]]) -> None:
        """Given a list of lists, appends None values to each list until the
        length of the longest list is matched."""
        self.__paths = list(map(Path, paths))
        self.validate_paths()
        self.normalize_path_matrix()

    def compact_path_matrix(self) -> None:
        """Eliminate trailing None columns in the path matrix."""
        while self.__paths and all(path[-1] is None for path in self.__paths):
            self.__paths = [path[:-1] for path in self.__paths]

    def delete_path(self, path: Path) -> None:
        """Deletes the specified path from the path matrix and leaf_slots"""
        idx = self.__paths.index(path)
        del self.__paths[idx]
        del self.leaf_slots[idx]
        self.compact_path_matrix()

    def __init__(self, path_strings: list[str], leaf_slots: list[MinimalSlot]):
        """Creates a new ModelGenerator"""
        paths = [path_string.split(".") for path_string in path_strings]
        self.load_path_strings(paths)
        self.load_leaf_slots(leaf_slots)


def update_metadata_model(model: MetadataModel, min_model: MinimalLinkMLModel) -> None:
    """Update a MetadataModel based on the provided minimal class representations.

    Bare bone slots will be added based on all slot names that are found in the
    provided classes. One class will be created according to each provided class
    with the slots being configured in slot_usage entirely.

    Args:
        model (MetadataModel): The MetadataModel to modify in place.
        class_names (dict[MinimalClass, str]): A dictionary of MinimalClasses
        with associated class names.
    """
    # Enumerate all unique slot names and add the slots to the model
    all_min_classes = itertools.chain(
        min_model.anonymous_classes, min_model.named_classes.values()
    )
    slot_names = sorted({slot.slot_name for cls in all_min_classes for slot in cls})

    for slot_name in slot_names:
        model.schema_view.add_slot(slot=SlotDefinition(name=slot_name))
    new_class_defs: dict[str, ClassDefinition] = {}

    for cls_name, min_cls_def in min_model.all_classes():
        new_class_defs[cls_name] = ClassDefinition(
            name=cls_name,
            slots=sorted(slot.slot_name for slot in min_cls_def),
            slot_usage={
                slot.slot_name: {
                    "range": slot.range,
                    "multivalued": slot.multivalued,
                }
                for slot in sorted(min_cls_def)
            },
        )

    all_classes = model.schema_view.all_classes()
    for cls_name, cls_def in new_class_defs.items():
        for slot_name, slot_config in cls_def.slot_usage.items():
            slot_range = slot_config["range"]
            if slot_range in all_classes or slot_range in new_class_defs:
                slot_config["inlined"] = True
                if slot_config["multivalued"]:
                    slot_config["inlined_as_list"] = True

    for cls_def in new_class_defs.values():
        model.schema_view.add_class(cls_def)


def add_aggregation(min_model: MinimalLinkMLModel, aggregation: Aggregation) -> None:
    """Adds the required slots and classes defined by the provided aggregation
    to the provided minimal model."""

    root_name = aggregation.output
    path_strings = [operation.output_path for operation in aggregation.operations]
    leaf_ranges = [op.function.result_range_name for op in aggregation.operations]
    leaf_multivalued = [op.function.result_multivalued for op in aggregation.operations]

    path_matrix = PathMatrix(
        path_strings=path_strings,
        leaf_slots=[
            MinimalSlot(range=slot_range, multivalued=mv)
            for slot_range, mv in zip(leaf_ranges, leaf_multivalued)
        ],
    )

    for _ in range(path_matrix.max_depth):
        classes: dict[Path, set[MinimalNamedSlot]] = defaultdict(set[MinimalNamedSlot])
        for idx, path_prefix in enumerate(path_matrix.paths):
            if path_prefix[-1]:
                classes[path_prefix[:-1]].add(path_matrix.leaf_slots[idx])
        for path_prefix, cls_set in classes.items():
            # Delete all paths with this prefix
            for slot in cls_set:
                path_matrix.delete_path(path_prefix + (slot.slot_name,))
            # Construct a class for this prefix and check if a matching
            # class already exists.
            cls = MinimalClass(cls_set)

            if not path_prefix:
                # We have reached the root of the output path model
                min_model.add_named_class(cls_def=cls, cls_name=root_name)
                slot_range = root_name
            else:
                # We are at an intermediate model and will use generic class
                # names and potentially re-use the classes as they fit
                if cls in min_model.anonymous_classes:
                    slot_range = min_model.anonymous_classes[cls]
                else:
                    slot_range = min_model.add_anonymous_class(cls)
            # Add the prefix as a new path with the identified class as range
            if path_prefix and path_prefix[-1]:
                path_matrix.add_path(
                    path_prefix,
                    MinimalNamedSlot(
                        slot_name=path_prefix[-1],
                        multivalued=False,
                        range=slot_range,
                    ),
                )


def bare_bone_model_from_other(model: MetadataModel) -> MetadataModel:
    """Create a new MetadataModel with no classes, enums, slots and subsets. All
    other SchemaDefinition attributes are kept."""
    model_dict = model.as_dict(essential=False)
    del model_dict["enums"]
    del model_dict["slots"]
    del model_dict["subsets"]
    model_dict["classes"] = {
        ROOT_CLASS: ClassDefinition(name=ROOT_CLASS, tree_root=True)
    }
    return MetadataModel(**model_dict)


def get_identifier_slot_names(
    input_model: MetadataModel, origin_map: dict[str, str]
) -> dict[str, str]:
    """Associate the correct identifier slot name with every output class.

    Args:
        input_model (MetadataModel): The input original metadata model
        origin_map (dict[str,str]): A dictionary mapping input class names to output class names

    Returns:
        dict[str,str]: A dictionary mapping output class names to identifier slot names
    """
    id_slot_map: dict[str, str] = {}
    for in_cls_name, out_cls_name in origin_map.items():
        slot_name = get_class_identifier(input_model, in_cls_name)
        if not slot_name:
            raise MetadataModelTransformationError(
                f"No identifier slot for class {in_cls_name}."
            )
        id_slot_map[out_cls_name] = slot_name
    return id_slot_map


def add_identifier_slots(
    input_model: MetadataModel,
    output_model: MetadataModel,
    origin_map: dict[str, str],
    id_slot_map: dict[str, str],
) -> MetadataModel:
    """Adds identifier slots to the high level output model classes. The slot
    name and range is inferred from the original input classes.

    Args:
        input_model (MetadataModel): _description_
        output_model (MetadataModel): _description_
        class_map (list[tuple[str, str]]): _description_

    Raises:
        MetadataModelTransformationError: _description_

    Returns:
        MetadataModel: _description_
    """
    output_schema_view = output_model.schema_view
    for in_class_name, out_class_name in origin_map.items():
        slot_name = id_slot_map[out_class_name]
        input_slot = input_model.schema_view.induced_slot(slot_name, in_class_name)
        output_schema_view = upsert_class_slot(
            schema_view=output_schema_view,
            class_name=out_class_name,
            new_slot=SlotDefinition(
                name=slot_name, range=input_slot.range, identifier=True
            ),
        )

    return output_schema_view.export_model()


def build_aggregation_model(
    model: MetadataModel, config: AggregateConfig
) -> MetadataModel:
    """Create a data model for the result of an aggregation transformation as
    described in an AggregateConfig

    Args:
        model (MetadataModel): The LinkML model to transform
        config (AggregateConfig): The config

    Returns:
        MetadataModel: A new, modified MetadataModel
    """
    # Reduce model to a bare bone model

    min_output_model = MinimalLinkMLModel()

    # Add function output classes
    for agg_func in (
        op.function for agg in config.aggregations for op in agg.operations
    ):
        cls_def = agg_func.result_range_cls_def
        cls_name = agg_func.result_range_name
        if cls_def:
            min_output_model.add_named_class(cls_def=cls_def, cls_name=cls_name)

    # Add upstream classes
    for aggregation in config.aggregations:
        add_aggregation(min_model=min_output_model, aggregation=aggregation)

    output_model = bare_bone_model_from_other(model)
    # Add slots and classes to the bare bone metadata model
    update_metadata_model(output_model, min_output_model)

    # Add identifier slots to classes
    origin_map = {agg.input: agg.output for agg in config.aggregations}
    id_slot_map = get_identifier_slot_names(input_model=model, origin_map=origin_map)
    output_model = add_identifier_slots(
        input_model=model,
        output_model=output_model,
        origin_map=origin_map,
        id_slot_map=id_slot_map,
    )

    # Anchor the output classes
    schema_view = output_model.schema_view
    for aggregation in config.aggregations:
        root_slot = snakecase(aggregation.output)
        root_slot = root_slot + "s" if not root_slot.endswith("s") else root_slot
        schema_view = add_anchor_point(
            schema_view=schema_view,
            anchor_point=AnchorPoint(
                target_class=aggregation.output,
                identifier_slot=id_slot_map[aggregation.output],
                root_slot=root_slot,
            ),
        )

    return schema_view.export_model()



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for checking transformation-specific model assumptions."""


from metldata.builtin_transformations.merge_slots.models import SlotMergeInstruction
from metldata.model_utils.assumptions import (
    MetadataModelAssumptionError,
    check_class_slot_exists,
)
from metldata.model_utils.essentials import MetadataModel


def check_source_slots_exist(
    model: MetadataModel, merge_instructions: list[SlotMergeInstruction]
) -> None:
    """Check that the source slots exist in the model."""

    for merge_instruction in merge_instructions:
        for source_slot in merge_instruction.source_slots:
            try:
                check_class_slot_exists(
                    model=model,
                    class_name=merge_instruction.class_name,
                    slot_name=source_slot,
                )
            except MetadataModelAssumptionError as error:
                raise MetadataModelAssumptionError(
                    f"Source slot {source_slot} of class "
                    + f"{merge_instruction.class_name} does not exist in the model."
                ) from error


def check_target_slots_not_exist(
    model: MetadataModel, merge_instructions: list[SlotMergeInstruction]
) -> None:
    """Check that the target slots do not exist in the model."""

    for merge_instruction in merge_instructions:
        try:
            check_class_slot_exists(
                model=model,
                class_name=merge_instruction.class_name,
                slot_name=merge_instruction.target_slot,
            )
        except MetadataModelAssumptionError:
            # this is expected
            continue
        else:
            raise MetadataModelAssumptionError(
                f"Target slot {merge_instruction.target_slot} of class "
                + f"{merge_instruction.class_name} already exists in the model."
            )


def check_model_class_slots(
    model: MetadataModel, merge_instructions: list[SlotMergeInstruction]
):
    """Check that the specified classes and slots exist in the model."""

    check_source_slots_exist(model=model, merge_instructions=merge_instructions)
    check_target_slots_not_exist(model=model, merge_instructions=merge_instructions)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pathlib import Path
from typing import Dict, List, Optional, Union

import yaml
from pydantic.env_settings import BaseSettings

from metldata.load.config import ArtifactLoaderAPIConfig
from tests.fixtures.load.utils import BASE_DIR

TEST_CONFIG_YAML = BASE_DIR / "test_config.yaml"


def get_config(
    sources: Optional[List[Union[BaseSettings, dict]]] = None,
    default_config_yaml: Path = TEST_CONFIG_YAML,
) -> ArtifactLoaderAPIConfig:
    """Merges parameters from the default TEST_CONFIG_YAML with params inferred
    from testcontainers."""
    sources_dict: Dict[str, object] = {}

    if sources is not None:
        for source in sources:
            if isinstance(source, BaseSettings):
                sources_dict.update(**source.dict())
            else:
                sources_dict.update(**source)

    with open(default_config_yaml, "r", encoding="UTF-8") as file:
        default_config = yaml.safe_load(file)

    return ArtifactLoaderAPIConfig(**default_config, **sources_dict)  # type: ignore



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""This module provides the CachedMetadataModel class."""

from metldata.model_utils.anchors import AnchorPoint, get_anchors_points_by_target
from metldata.model_utils.essentials import MetadataModel


class CachedMetadataModel:
    """A model for a MetadataModel and pre-computed, non-dynamic associated
    anchor points by target and class names."""

    model: MetadataModel
    anchors_points_by_target: dict[str, AnchorPoint]
    all_class_names: list[str]

    def __init__(self, model: MetadataModel):
        self.model = model
        self.anchors_points_by_target = get_anchors_points_by_target(model=model)
        self.all_classes = list(model.schema_view.all_classes())



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the metadata utils."""


import pytest

from metldata.metadata_utils import (
    MetadataResourceNotFoundError,
    convert_resource_list_to_dict,
    get_resources_of_class,
    lookup_resource_by_identifier,
    upsert_resources_in_metadata,
)
from metldata.model_utils.anchors import AnchorPoint, get_anchors_points_by_target
from tests.fixtures.metadata import VALID_MINIMAL_METADATA_EXAMPLE
from tests.fixtures.metadata_models import VALID_MINIMAL_METADATA_MODEL


def test_lookup_resource_by_identifier_happy():
    """Test the happy path of using the lookup_resource_by_identifier function."""

    identifier = "test_sample_01_R1"
    resource_dict = convert_resource_list_to_dict(
        resources=VALID_MINIMAL_METADATA_EXAMPLE["files"], identifier_slot="alias"
    )
    expected_resource = resource_dict[identifier]

    anchor_points_by_target = get_anchors_points_by_target(
        model=VALID_MINIMAL_METADATA_MODEL
    )
    observed_resource = lookup_resource_by_identifier(
        class_name="File",
        global_metadata=VALID_MINIMAL_METADATA_EXAMPLE,
        identifier=identifier,
        anchor_points_by_target=anchor_points_by_target,
    )

    assert observed_resource == expected_resource


def test_lookup_resource_by_identifier_not_exist():
    """Test the using the lookup_resource_by_identifier function with an identifier
    that does not exist."""

    identifier = "non_existing_identifier"

    anchor_points_by_target = get_anchors_points_by_target(
        model=VALID_MINIMAL_METADATA_MODEL
    )

    with pytest.raises(MetadataResourceNotFoundError):
        _ = lookup_resource_by_identifier(
            class_name="File",
            global_metadata=VALID_MINIMAL_METADATA_EXAMPLE,
            identifier=identifier,
            anchor_points_by_target=anchor_points_by_target,
        )


EXAMPLE_GLOBAL_METADATA = {
    "files": [
        {
            "alias": "test_sample_01_R1",
            "file_format": "fastq",
            "file_size": 1234567890,
        },
        {
            "alias": "test_sample_01_R2",
            "file_format": "fastq",
            "file_size": 1234567890,
        },
    ]
}

EXAMPLE_ANCHOR_POINTS_BY_TARGET = {
    "File": AnchorPoint(target_class="File", identifier_slot="alias", root_slot="files")
}


def test_get_resources_of_class_happy():
    """Test the happy path of using the get_resources_of_class function."""

    expected_resources = EXAMPLE_GLOBAL_METADATA["files"]

    observed_resources = get_resources_of_class(
        class_name="File",
        global_metadata=EXAMPLE_GLOBAL_METADATA,
        anchor_points_by_target=EXAMPLE_ANCHOR_POINTS_BY_TARGET,
    )

    assert observed_resources == expected_resources


def test_update_resources_in_metadata_happy():
    """Test the happy path of using the update_resources_in_metadata function."""

    modified_resources = [
        {
            "alias": "test_sample_01_R1",
            "file_format": "fastq",
            "some_new_field": "some_new_value",
        },
        {
            "alias": "test_sample_01_R2",
            "file_format": "fastq",
            "some_new_field": "some_new_value",
        },
    ]

    expected_metadata = {"files": modified_resources}

    observed_metadata = upsert_resources_in_metadata(
        resources=modified_resources,
        class_name="File",
        global_metadata=EXAMPLE_GLOBAL_METADATA,
        anchor_points_by_target=EXAMPLE_ANCHOR_POINTS_BY_TARGET,
    )

    assert observed_metadata == expected_metadata



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the main module."""

import json

import pytest

from metldata.event_handling.artifact_events import get_artifact_topic
from metldata.event_handling.models import SubmissionEventPayload
from metldata.transform.main import (
    TransformationEventHandlingConfig,
    run_workflow_on_all_source_events,
)
from tests.fixtures.event_handling import file_system_event_fixture  # noqa: F401
from tests.fixtures.event_handling import Event, FileSystemEventFixture
from tests.fixtures.workflows import (
    EXAMPLE_WORKFLOW_DEFINITION,
    EXAMPLE_WORKFLOW_TEST_CASE,
)


@pytest.mark.asyncio
async def test_run_workflow_on_all_source_events(
    file_system_event_fixture: FileSystemEventFixture,  # noqa: F811
):
    """Test the happy path of using the run_workflow_on_all_source_events function."""

    event_config = TransformationEventHandlingConfig(
        artifact_topic_prefix="artifacts",
        source_event_topic="source-events",
        source_event_type="source-event",
        **file_system_event_fixture.config.dict(),
    )

    submission_id = "some-submission-id"
    source_event = Event(
        topic=event_config.source_event_topic,
        type_=event_config.source_event_type,
        key=submission_id,
        payload=json.loads(
            SubmissionEventPayload(
                submission_id=submission_id,
                content=EXAMPLE_WORKFLOW_TEST_CASE.original_metadata,
                annotation=EXAMPLE_WORKFLOW_TEST_CASE.submission_annotation,
            ).json()
        ),
    )
    await file_system_event_fixture.publish_events(events=[source_event])

    expected_events = [
        Event(
            topic=get_artifact_topic(
                artifact_topic_prefix=event_config.artifact_topic_prefix,
                artifact_type=artifact_type,
            ),
            type_=artifact_type,
            key=submission_id,
            payload=json.loads(
                SubmissionEventPayload(
                    submission_id=submission_id,
                    content=artifact,
                    annotation=EXAMPLE_WORKFLOW_TEST_CASE.submission_annotation,
                ).json()
            ),
        )
        for artifact_type, artifact in EXAMPLE_WORKFLOW_TEST_CASE.artifact_metadata.items()
    ]

    await run_workflow_on_all_source_events(
        event_config=event_config,
        workflow_definition=EXAMPLE_WORKFLOW_DEFINITION,
        worflow_config=EXAMPLE_WORKFLOW_TEST_CASE.config,
        original_model=EXAMPLE_WORKFLOW_TEST_CASE.original_model,
    )

    file_system_event_fixture.expect_events(expected_events=expected_events)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling submissions."""


from ghga_service_commons.utils.utc_dates import now_as_utc

from metldata.accession_registry.accession_registry import AccessionRegistry
from metldata.custom_types import SubmissionContent
from metldata.model_utils.anchors import get_anchors_points_by_target
from metldata.model_utils.config import MetadataModelConfig
from metldata.model_utils.metadata_validator import MetadataValidator
from metldata.submission_registry import models
from metldata.submission_registry.event_publisher import SourceEventPublisher
from metldata.submission_registry.identifiers import (
    generate_accession_map,
    generate_submission_id,
)
from metldata.submission_registry.submission_store import SubmissionStore


class SubmissionRegistryConfig(MetadataModelConfig):
    """Config parameters and their defaults."""


class SubmissionRegistry:
    """A class for handling submissions."""

    # error shortcuts:
    SubmissionDoesNotExistError = SubmissionStore.SubmissionDoesNotExistError
    ValidationError = MetadataValidator.ValidationError

    class StatusError(RuntimeError):
        """Raised when the requested action requires a different submission status."""

        def __init__(
            self, *, submission_id: str, current_status: str, expected_status: str
        ):
            message = (
                f"The submission with id '{submission_id} has the status"
                + f" '{current_status}' but expected '{expected_status}'."
            )
            super().__init__(message)

    class ContentEmptyError(RuntimeError):
        """Raised when the content of a submission is empty.."""

        def __init__(self, *, submission_id: str):
            message = (
                f"The content for the submission with id '{submission_id} has not yet"
                + " been specified."
            )
            super().__init__(message)

    def __init__(
        self,
        *,
        config: SubmissionRegistryConfig,
        submission_store: SubmissionStore,
        event_publisher: SourceEventPublisher,
        accession_registry: AccessionRegistry,
    ):
        """Initialize with dependencies and config parameters."""

        self._submission_store = submission_store
        self._metadata_validator = MetadataValidator(model=config.metadata_model)
        self._event_publisher = event_publisher
        self._accession_registry = accession_registry
        self._anchor_points_by_target = get_anchors_points_by_target(
            model=config.metadata_model
        )

    def _get_submission_with_status(
        self, *, id_: str, expected_status: models.SubmissionStatus
    ) -> models.Submission:
        """Get details for a submission that is assumed to have the specified status.

        Raises:
            SubmissionRegistry.SubmissionDoesNotExistError:
                when no submission with the specified ID exists.
            SubmissionRegistry.StatusError:
                when the submission does not have the expected status.
        """

        # raises SubmissionDoesNotExistError if failed to get submission by id:
        submission = self._submission_store.get_by_id(submission_id=id_)

        if submission.current_status != expected_status:
            raise self.StatusError(
                submission_id=id_,
                current_status=str(submission.current_status),
                expected_status=str(expected_status),
            )

        return submission

    def init_submission(self, *, header: models.SubmissionHeader) -> str:
        """Initialize a new submission by providing header information. Returns the
        ID of the created submission. No source event is published, yet, since the
        submission content is still empty."""

        id_ = generate_submission_id()
        submission_creation = models.Submission(id=id_, **header.dict())
        self._submission_store.insert_new(submission=submission_creation)

        return id_

    def get_submission(self, *, id_: str) -> models.Submission:
        """Get details on the existing submission with the specified id.

        Raises:
            SubmissionRegistry.SubmissionDoesNotExistError:
                when no submission with the specified ID exists.
        """

        return self._submission_store.get_by_id(id_)

    def upsert_submission_content(
        self, *, submission_id: str, content: SubmissionContent
    ) -> None:
        """Insert or update the content of a pending submission.
        The metadata is validated against the model, persisted in the submission store,
        and finally published as source event.

        Raises:
            SubmissionRegistry.SubmissionDoesNotExistError:
                when no submission exists for the specified id.
            SubmissionRegistry.ValidationError:
                when the provided content failed validation against the metadata model.
            SubmissionRegistry.StatusError:
                when the status of the specified submission is not pending.
        """

        submission = self._get_submission_with_status(
            id_=submission_id,
            expected_status=models.SubmissionStatus.PENDING,
        )

        # raises ValidationError if not valid:
        self._metadata_validator.validate(content)

        updated_accession_map = generate_accession_map(
            content=content,
            existing_accession_map=submission.accession_map,
            accession_registry=self._accession_registry,
            anchor_points_by_target=self._anchor_points_by_target,
        )

        updated_submission = submission.copy(
            update={"content": content, "accession_map": updated_accession_map}
        )
        self._submission_store.update_existing(submission=updated_submission)

        self._event_publisher.publish_submission(updated_submission)

    def complete_submission(self, *, id_: str) -> None:
        """Declare an existing submission as complete. The content of the submission
        cannot change anymore afterwards.


        Raises:
            SubmissionRegistry.SubmissionDoesNotExistError:
                when no submission exists for the specified id.
            SubmissionRegistry.StatusError:
                when the status of the specified submission is not pending.
            SubmissionRegistry.ContentEmptyError:
                when the content of the specified sumbission has not yet been specified.
        """

        submission = self._get_submission_with_status(
            id_=id_, expected_status=models.SubmissionStatus.PENDING
        )

        if submission.content is None:
            raise self.ContentEmptyError(submission_id=id_)

        status_change = models.StatusChange(
            timestamp=now_as_utc(), new_status=models.SubmissionStatus.COMPLETED
        )
        updated_submission = submission.copy(
            update={"status_history": submission.status_history + (status_change,)}
        )
        self._submission_store.update_existing(submission=updated_submission)

        self._event_publisher.publish_submission(updated_submission)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Data models"""

from enum import Enum
from operator import attrgetter
from typing import Optional

from ghga_service_commons.utils.utc_dates import DateTimeUTC, now_as_utc
from pydantic import BaseModel, Field, validator
from typing_extensions import TypeAlias

from metldata.custom_types import SubmissionContent

AccessionMap: TypeAlias = dict[str, dict[str, str]]


class SubmissionStatus(Enum):
    """Statuses a submission may have."""

    PENDING = "pending"
    IN_REVIEW = "in-review"
    CANCELED = "canceled"
    COMPLETED = "completed"
    DEPRECATED_PREPUBLICATION = "deprecated-prepublication"
    EMPTIED_PREPUBLICATION = "emptied-prepublication"
    PUBLISHED = "published"
    DEPRECATED_POSTPUBLICATION = "deprecated-postpublication"
    HIDDEN_POSTPUBLICATION = "hidden-postpublication"
    EMPTIED_POSTPUBLICATION = "emptied-postpublication"


class StatusChange(BaseModel):
    """A model for describing status changes of submissions."""

    timestamp: DateTimeUTC
    new_status: SubmissionStatus


class SubmissionHeader(BaseModel):
    """Basic information provided for a submission."""

    title: str = Field(..., description="A descriptive title for this submission.")
    description: Optional[str] = Field(None, description="An optional description.")


class Submission(SubmissionHeader):
    """A model for describing a submission."""

    id: str

    content: Optional[SubmissionContent] = Field(
        None,
        description=(
            "The metadata content of the submission. Keys on the top level correspond to"
            + " names of anchored metadata classes. Keys and values on the second level"
            + " correspond to the user-defined aliases and contents of class instance. Please note,"
            + " that the user-defined alias might only be unique within the scope of"
            + " the coressponding class and this submission."
        ),
    )

    accession_map: AccessionMap = Field(
        default_factory=dict,
        description=(
            "A map of user-specified id to system-generated accession for metadata"
            + " resources. Keys on the top level correspond to names of metadata classes."
            + " Keys on the second level correspond to user-specified aliases."
            + " Values on the second level correspond to system-generated accessions."
            + " Please note, that the user-defined alias might only be unique within"
            + " the scope of the coressponding class and this submission. By contrast,"
            + " the system-generated accession is unique across all classes and"
            + " submissions."
        ),
    )

    status_history: tuple[StatusChange, ...] = Field(
        default_factory=lambda: (
            StatusChange(timestamp=now_as_utc(), new_status=SubmissionStatus.PENDING),
        ),
        description="A history of status changes.",
    )

    @property
    def current_status(self) -> SubmissionStatus:
        """Extract the current submission status from the status history."""

        if len(self.status_history) == 0:
            raise RuntimeError("Status history is empty.")

        sorted_history = sorted(self.status_history, key=attrgetter("timestamp"))

        return sorted_history[-1].new_status

    # pylint: disable=no-self-argument
    @validator("accession_map")
    def check_accession_uniqueness(cls, value: AccessionMap) -> AccessionMap:
        """Check that no accessions are re-used accross classes."""

        total_resources = 0
        all_accessions: set[str] = set()

        for resources in value.values():
            total_resources += len(resources)
            all_accessions.update(resources.values())

        if len(all_accessions) != total_resources:
            raise ValueError("Accessions are not unique.")

        return value



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Data models"""

from enum import Enum

from pydantic import BaseModel, Field


class ReferencePathElementType(Enum):
    """The type of ReferencePathElements.

    Can be active, meaning the source class is referencing the target class using the
    specified slot.
    Or passive, meaning that the source class is referenced by the target class and the
    slot is part of the target class.
    """

    ACTIVE = "active"
    PASSIVE = "passive"


class ReferencePathElement(BaseModel):
    """A model describing an element of a reference path between classes of a
    metadata model as further explained by the ReferencePath.
    """

    type_: ReferencePathElementType = Field(
        ...,
        description=(
            "The type of reference. Active or passive as explained in the"
            + " ReferencePathElementType enum."
        ),
    )
    source: str = Field(
        ..., description="The name of the source class that is referencing."
    )
    target: str = Field(
        ..., description="The name of the target class that is referenced."
    )
    slot: str = Field(
        ...,
        description=(
            "The name of the slot that holds the reference."
            + " In case of a active type, the slot is part of the source class."
            + " In case of a passive type, the slot is part of the target class."
        ),
    )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config parameters and their defaults."""

from pydantic import BaseSettings, Field


class AccessionAdditionConfig(BaseSettings):
    """Config to add accessions to a model and associated metadata."""

    accession_slot_name: str = Field(
        "accession", description="The name of the slot to contain the accessions to."
    )
    accession_slot_description: str = Field(
        "The accession for an entity.",
        description="The description of the slot to contain the accessions to.",
    )

    class Config:
        """Pydantic config."""

        extra = "forbid"



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Functionality for collecting available artifacts."""

from collections import defaultdict

from pydantic import Field, validator

from metldata.event_handling.artifact_events import (
    ArtifactEventConfig,
    get_artifact_topic,
)
from metldata.event_handling.event_handling import FileSystemEventCollector
from metldata.load.models import ArtifactResourceDict


class ArtifactCollectorConfig(ArtifactEventConfig):
    """Config parameters and their defaults."""

    artifact_types: list[str] = Field(
        ...,
        description=(
            "The artifacts types of interest. Together with the topic prefix, they"
            + " determine the topics to subscribe to. The artifact types must not"
            + " contain dots."
        ),
    )

    # pylint: disable=no-self-argument
    @validator("artifact_types")
    def artifact_types_must_not_contain_dots(cls, value: list[str]):
        """Validate that artifact types do not contain dots."""

        for artifact_type in value:
            if "." in artifact_type:
                raise ValueError(
                    f"Artifact type '{artifact_type}' must not contain dots."
                )

        return value


def collect_artifacts(
    *, config: ArtifactCollectorConfig, event_collector: FileSystemEventCollector
) -> ArtifactResourceDict:
    """Collect artifacts from the file system."""

    artifact_resources: ArtifactResourceDict = defaultdict(list)
    for artifact_type in config.artifact_types:
        topic = get_artifact_topic(
            artifact_topic_prefix=config.artifact_topic_prefix,
            artifact_type=artifact_type,
        )
        for event in event_collector.collect_events(topic=topic):
            content = event.payload.get("content")
            if not content:
                raise RuntimeError("Artifact does not contain 'content' field.")
            artifact_resources[artifact_type].append(content)

    return artifact_resources



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for authorization using alphanumeric tokens."""

import secrets
from hashlib import sha256


def generate_token() -> str:
    """Generate a random token."""
    return secrets.token_urlsafe(32)


def hash_token(token: str) -> str:
    """Hash the given token."""

    return sha256(token.encode()).hexdigest()


def generate_token_and_hash() -> tuple[str, str]:
    """Generate a random token and its hash."""

    token = generate_token()
    token_hash = hash_token(token)
    return token, token_hash


def check_token(token: str, token_hashes: list[str]) -> bool:
    """Check whether the given token matches one of the given token hashes."""

    return hash_token(token) in token_hashes



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Config Parameter Modeling and Parsing"""

from pathlib import Path

from pydantic import BaseSettings, Field

from metldata.model_utils.assumptions import check_basic_model_assumption
from metldata.model_utils.essentials import MetadataModel


class MetadataModelConfig(BaseSettings):
    """Config parameters and their defaults."""

    metadata_model_path: Path = Field(
        ..., description="The path to the metadata model defined in LinkML."
    )

    @property
    def metadata_model(self) -> MetadataModel:
        """Load the model from the path and check basic assumptions."""

        model = MetadataModel.init_from_path(self.metadata_model_path)
        check_basic_model_assumption(model)
        return model



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test the config module."""

from contextlib import nullcontext

import pytest
from pydantic import ValidationError

from metldata.builtin_transformations.merge_slots.config import SlotMergingConfig


@pytest.mark.parametrize(
    "merge_instructions, valid",
    [
        (
            # valid since two different classes:
            [
                {
                    "class_name": "class_a",
                    "source_slots": ["some_slot", "another_slot"],
                    "target_slot": "merged_slot",
                },
                {
                    "class_name": "class_b",
                    "source_slots": ["merged_slot", "yet_another_slot"],
                    "target_slot": "some_slot",
                },
            ],
            True,
        ),
        (
            # invalid since using target slot of one merge instruction as source slot
            # for another:
            [
                {
                    "class_name": "class_a",
                    "source_slots": ["some_slot", "another_slot"],
                    "target_slot": "merged_slot",
                },
                {
                    "class_name": "class_a",
                    "source_slots": ["merged_slot", "yet_another_slot"],
                    "target_slot": "another_merged_slot",
                },
            ],
            False,
        ),
        (
            # invalid since two merge instruction target the same slot:
            [
                {
                    "class_name": "class_a",
                    "source_slots": ["some_slot", "another_slot"],
                    "target_slot": "merged_slot",
                },
                {
                    "class_name": "class_a",
                    "source_slots": ["another_slot", "yet_another_slot"],
                    "target_slot": "merged_slot",
                },
            ],
            False,
        ),
    ],
)
def test_slot_merging_config(merge_instructions: list, valid: bool):
    """Test that validation of SlotMergingConfig."""

    with nullcontext() if valid else pytest.raises(ValidationError):  # type: ignore
        SlotMergingConfig(merge_instructions=merge_instructions)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Fixtures for handling events."""

from dataclasses import dataclass
from pathlib import Path

import pytest

from metldata.event_handling.event_handling import (
    Event,
    FileSystemEventCollector,
    FileSystemEventConfig,
    FileSystemEventPublisher,
)


class EventExpectationMismatch(RuntimeError):
    """Raised when expected events where not found."""

    def __init__(self, expected_events: set[str], consumed_events: set[str]):
        message = f"Expected events '{expected_events}' but got '{consumed_events}'"
        super().__init__(message)


@dataclass
class FileSystemEventFixture:
    """Returned by file_system_event_fixture."""

    config: FileSystemEventConfig
    publisher: FileSystemEventPublisher
    collector: FileSystemEventCollector

    def expect_events(self, expected_events: list[Event]) -> None:
        """Check if the events expected to be published can be consumed.

        Raises:
            EventExpectationMismatch: If the expected events are not consumed.
        """

        topics = sorted({event.topic for event in expected_events})
        types = sorted({event.type_ for event in expected_events})

        observed_events: list[Event] = []
        for topic in topics:
            observed_events.extend(
                self.collector.collect_events(topic=topic, types=types)
            )

        # hashable versions for comparison:
        observed_event_jsons = {event.json() for event in observed_events}
        expected_event_jsons = {event.json() for event in expected_events}

        if expected_event_jsons != observed_event_jsons:
            raise EventExpectationMismatch(
                expected_events=observed_event_jsons,
                consumed_events=expected_event_jsons,
            )

    async def publish_events(self, events: list[Event]) -> None:
        """Publish a list of events."""

        for event in events:
            await self.publisher.publish(
                topic=event.topic,
                type_=event.type_,
                key=event.key,
                payload=event.payload,
            )


@pytest.fixture
def file_system_event_fixture(tmp_path: Path) -> FileSystemEventFixture:
    """A fixture for handling events on the file system."""

    config = FileSystemEventConfig(
        event_store_path=tmp_path,
    )
    publisher = FileSystemEventPublisher(config=config)
    collector = FileSystemEventCollector(config=config)
    return FileSystemEventFixture(
        config=config, publisher=publisher, collector=collector
    )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config for aggregate transformations"""

from typing import Optional

from pydantic import BaseModel, BaseSettings, root_validator

from metldata.builtin_transformations.aggregate.func import (
    AggregationFunction,
    transformation_by_name,
)


class AggregationOperation(BaseModel):
    """A model for a single aggregation operation executed on one or multiple
    branches in the data described by a path in the model."""

    input_paths: list[str]
    output_path: str
    visit_only_once: Optional[list[str]] = None
    function: type[AggregationFunction]

    # pylint: disable=no-self-argument
    @root_validator(pre=True)
    def lookup_operation(cls, values: dict) -> dict:
        """Replaces operation strings with operation types."""
        if "function" in values:
            values["function"] = transformation_by_name(values["function"])
        # not raising an error otherwise as pydantic will do that in following
        # validation
        return values


class Aggregation(BaseModel):
    """Model for an aggregation."""

    input: str
    output: str
    operations: list[AggregationOperation]


class AggregateConfig(BaseSettings):
    """A model for the configuration of the aggregate transformation."""

    aggregations: list[Aggregation]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Metadata transformation functionality for the aggregate transformation."""

from metldata.builtin_transformations.aggregate.cached_model import CachedMetadataModel
from metldata.builtin_transformations.aggregate.config import Aggregation
from metldata.builtin_transformations.aggregate.data_subgraph import DataSubgraph
from metldata.builtin_transformations.aggregate.expanding_dict import ExpandingDict
from metldata.builtin_transformations.aggregate.func import MetadataTransformationError
from metldata.custom_types import Json
from metldata.model_utils.anchors import AnchorPoint


# pylint: disable=unused-argument
def execute_aggregation(
    *,
    original_model: CachedMetadataModel,
    original_data: Json,
    aggregation: Aggregation,
) -> list[Json]:
    """Transforms the metadata according to the specified aggregation operation."""
    anchor_point = original_model.anchors_points_by_target[aggregation.input]
    id_slot = anchor_point.identifier_slot
    input_anchor_data = original_data[anchor_point.root_slot]
    output_data: list[Json] = []
    for input_element in input_anchor_data:
        result = ExpandingDict()
        for operation in aggregation.operations:
            subgraph = DataSubgraph(
                model=original_model,
                submission_data=original_data,
                origin=aggregation.input,
                path_strings=operation.input_paths,
                visit_once_classes=operation.visit_only_once,
            )
            try:
                aggregated = operation.function.func(
                    subgraph.terminal_nodes(data=input_element)
                )
            except Exception as error:
                raise MetadataTransformationError(
                    "Cannot execute operation:\n"
                    f"{operation}\nwith input {input_element!r}:\n{error}"
                ) from error
            result.set_path_value(operation.output_path, aggregated)
        result[id_slot] = input_element[id_slot]
        output_data.append(result.to_dict())

    return output_data


def execute_aggregations(
    *,
    original_model: CachedMetadataModel,
    transformed_anchors_points: dict[str, AnchorPoint],
    metadata: Json,
    aggregations: list[Aggregation],
) -> Json:
    """Transforms the metadata according to the specified list of aggregation
    operations."""
    transformed_data = {}
    for aggregation in aggregations:
        output_data = execute_aggregation(
            original_model=original_model,
            original_data=metadata,
            aggregation=aggregation,
        )
        transformed_data[
            transformed_anchors_points[aggregation.output].root_slot
        ] = output_data
    return transformed_data



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A GHGA Archive-specific workflow."""

from metldata.builtin_transformations.add_accessions import (
    ACCESSION_ADDITION_TRANSFORMATION,
)
from metldata.builtin_transformations.aggregate import AGGREGATE_TRANSFORMATION
from metldata.builtin_transformations.custom_embeddings import (
    CUSTOM_EMBEDDING_TRANSFORMATION,
)
from metldata.builtin_transformations.delete_slots import SLOT_DELETION_TRANSFORMATION
from metldata.builtin_transformations.infer_references import (
    REFERENCE_INFERENCE_TRANSFORMATION,
)
from metldata.builtin_transformations.merge_slots import SLOT_MERGING_TRANSFORMATION
from metldata.builtin_transformations.normalize_model import (
    NORMALIZATION_TRANSFORMATION,
)
from metldata.transform.base import WorkflowDefinition, WorkflowStep

GHGA_ARCHIVE_WORKFLOW = WorkflowDefinition(
    description=("A GHGA Archive-specific workflow"),
    steps={
        "normalize_model": WorkflowStep(
            description=("Transform the model to a canonical form."),
            transformation_definition=NORMALIZATION_TRANSFORMATION,
            input=None,
        ),
        "add_accessions": WorkflowStep(
            description="Add accessions that replace aliases as identifiers.",
            transformation_definition=ACCESSION_ADDITION_TRANSFORMATION,
            input="normalize_model",
        ),
        "embed_restricted": WorkflowStep(
            description=(
                "A step to generate a fully embedded metadata for data recipients that"
                + " have been granted controlled access to that dataset."
            ),
            transformation_definition=CUSTOM_EMBEDDING_TRANSFORMATION,
            input="add_accessions",
        ),
        "infer_multiway_references": WorkflowStep(
            description=(
                "Infer multi-way references. During submission, references between"
                + " two classes are only establish in a uni-lateral way from one of the"
                + " interlinked classes. Here references from both sides are"
                + " established. Moreover, shortcuts between entities that are"
                + " connected via a common reference are established."
            ),
            transformation_definition=REFERENCE_INFERENCE_TRANSFORMATION,
            input="add_accessions",
        ),
        "merge_dataset_file_lists": WorkflowStep(
            description=(
                "Generate a slot in that dataset that combines all files of different"
                + " type into a single list."
            ),
            transformation_definition=SLOT_MERGING_TRANSFORMATION,
            input="infer_multiway_references",
        ),
        "remove_restricted_metadata": WorkflowStep(
            description=(
                "Restricted metadata that shall not be publicly accessible is removed."
            ),
            transformation_definition=SLOT_DELETION_TRANSFORMATION,
            input="merge_dataset_file_lists",
        ),
        "aggregate_stats": WorkflowStep(
            description="Compute aggregate statistics.",
            transformation_definition=AGGREGATE_TRANSFORMATION,
            input="remove_restricted_metadata",
        ),
        "embed_public": WorkflowStep(
            description=(
                "A step to generate a fully embedded metadata resources for use in the"
                + " metadata catalog as the single dataset view."
            ),
            transformation_definition=CUSTOM_EMBEDDING_TRANSFORMATION,
            input="remove_restricted_metadata",
        ),
    },
    artifacts={
        "embedded_restricted": "embed_restricted",
        "resolved_restricted": "merge_dataset_file_lists",
        "resolved_public": "remove_restricted_metadata",
        "embedded_public": "embed_public",
        "stats_public": "aggregate_stats",
    },
)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling Transformation."""

from pydantic import BaseModel

from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.essentials import MetadataModel
from metldata.model_utils.metadata_validator import MetadataValidator
from metldata.transform.base import (
    Config,
    TransformationDefinition,
    WorkflowConfig,
    WorkflowDefinition,
    WorkflowStep,
    WorkflowStepBase,
)


class WorkflowConfigMismatchError(RuntimeError):
    """Raised when the provided workflow config does not match the config class of the
    workflow definition.
    """

    def __init__(
        self, workflow_definition: WorkflowDefinition, workflow_config: Config
    ):
        """Initialize the error with the workflow definition and the config."""

        message = (
            f"The config {workflow_config} is not an instance of the config class "
            f"{workflow_definition.config_cls} of the workflow definition "
            f"{workflow_definition}."
        )
        super().__init__(message)


class TransformationHandler:
    """Used for executing transformations described in a TransformationDefinition."""

    def __init__(
        self,
        transformation_definition: TransformationDefinition[Config],
        transformation_config: Config,
        original_model: MetadataModel,
    ):
        """Initialize the TransformationHandler by checking the assumptions made on the
        original model and transforming the model as described in the transformation
        definition. The transformed model is available at the `transformed_model`
        attribute.

        Raises:
            ModelAssumptionError:
                if the assumptions made on the original model are not met.
        """

        self._definition = transformation_definition
        self._config = transformation_config
        self._original_model = original_model

        self._definition.check_model_assumptions(self._original_model, self._config)
        self.transformed_model = self._definition.transform_model(
            self._original_model, self._config
        )
        self._metadata_transformer = self._definition.metadata_transformer_factory(
            config=self._config,
            original_model=self._original_model,
            transformed_model=self.transformed_model,
        )

        self._original_metadata_validator = MetadataValidator(
            model=self._original_model
        )
        self._transformed_metadata_validator = MetadataValidator(
            model=self.transformed_model
        )

    def transform_metadata(
        self, metadata: Json, *, annotation: SubmissionAnnotation
    ) -> Json:
        """Transforms metadata using the transformation definition. Validates the
        original metadata against the original model and the transformed metadata
        against the transformed model.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """

        self._original_metadata_validator.validate(metadata)
        transformed_metadata = self._metadata_transformer.transform(
            metadata=metadata, annotation=annotation
        )
        self._transformed_metadata_validator.validate(transformed_metadata)

        return transformed_metadata


class ResolvedWorkflowStep(WorkflowStepBase):
    """A resolved workflow step contains a transformation handler."""

    transformation_handler: TransformationHandler

    class Config:
        """Config for ResolvedWorkflowStep."""

        arbitrary_types_allowed = True


class ResolvedWorkflow(WorkflowDefinition):
    """A resolved workflow contains a list of resolved workflow steps."""

    steps: dict[str, ResolvedWorkflowStep]  # type: ignore
    workflow_config: WorkflowConfig


def check_workflow_config(
    *, workflow_definition: WorkflowDefinition, workflow_config: WorkflowConfig
):
    """Checks if the config is an instance of the config class of the workflow
    definition.

    Raises:
        WorkflowConfigMismatchError:
    """

    if workflow_config.schema_json() == workflow_definition.schema_json():
        raise WorkflowConfigMismatchError(
            workflow_definition=workflow_definition, workflow_config=workflow_config
        )


def resolve_workflow_step(
    *,
    workflow_step: WorkflowStep,
    step_name: str,
    workflow_definition: WorkflowDefinition,
    workflow_config: WorkflowConfig,
    original_model: MetadataModel,
) -> ResolvedWorkflowStep:
    """Translates a workflow step given a workflow definition and a workflow config
    into a resolved workflow step.
    """

    check_workflow_config(
        workflow_definition=workflow_definition, workflow_config=workflow_config
    )

    transformation_config: BaseModel = getattr(workflow_config, step_name)
    transformation_handler = TransformationHandler(
        transformation_definition=workflow_step.transformation_definition,
        transformation_config=transformation_config,
        original_model=original_model,
    )
    return ResolvedWorkflowStep(
        transformation_handler=transformation_handler,
        input=workflow_step.input,
        description=workflow_step.description,
    )


def resolve_workflow(
    workflow_definition: WorkflowDefinition,
    original_model: MetadataModel,
    workflow_config: WorkflowConfig,
) -> ResolvedWorkflow:
    """Translates a workflow definition given an input model and a workflow config into
    a resolved workflow.
    """

    check_workflow_config(
        workflow_definition=workflow_definition, workflow_config=workflow_config
    )

    resolved_steps: dict[str, ResolvedWorkflowStep] = {}
    for step_name in workflow_definition.step_order:
        workflow_step = workflow_definition.steps[step_name]
        input_model = (
            original_model
            if workflow_step.input is None
            else resolved_steps[
                workflow_step.input
            ].transformation_handler.transformed_model
        )

        resolved_steps[step_name] = resolve_workflow_step(
            workflow_step=workflow_step,
            step_name=step_name,
            workflow_definition=workflow_definition,
            workflow_config=workflow_config,
            original_model=input_model,
        )

    return ResolvedWorkflow(
        steps=resolved_steps,
        workflow_config=workflow_config,
        description=workflow_definition.description,
        artifacts=workflow_definition.artifacts,
    )


def get_model_artifacts_from_resolved_workflow(resolved_workflow: ResolvedWorkflow):
    """Returns a dictionary of models for artifacts produced by resolved workflow."""

    return {
        artifact_name: resolved_workflow.steps[
            step_name
        ].transformation_handler.transformed_model
        for artifact_name, step_name in resolved_workflow.artifacts.items()
    }


class WorkflowHandler:
    """Used for executing workflows described in a WorkflowDefinition."""

    def __init__(
        self,
        workflow_definition: WorkflowDefinition,
        workflow_config: WorkflowConfig,
        original_model: MetadataModel,
    ):
        """Initialize the WorkflowHandler with a workflow deinition, a matching
        config, and a metadata model. The workflow definition is translated into a
        resolved workflow.
        """

        self._resolved_workflow = resolve_workflow(
            workflow_definition=workflow_definition,
            original_model=original_model,
            workflow_config=workflow_config,
        )

        self.artifact_models = get_model_artifacts_from_resolved_workflow(
            self._resolved_workflow
        )

    def run(
        self, *, metadata: Json, annotation: SubmissionAnnotation
    ) -> dict[str, Json]:
        """Run the workflow definition on metadata and its annotation to generate
        artifacts."""

        transformed_metadata: dict[str, Json] = {}
        for step_name in self._resolved_workflow.step_order:
            step = self._resolved_workflow.steps[step_name]
            input_metadata = (
                metadata if step.input is None else transformed_metadata[step.input]
            )
            transformed_metadata[
                step_name
            ] = step.transformation_handler.transform_metadata(
                input_metadata, annotation=annotation
            )

        return {
            artifact_name: transformed_metadata[step_name]
            for artifact_name, step_name in self._resolved_workflow.artifacts.items()
        }



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Test reference utils."""

from contextlib import nullcontext
from typing import Optional

import pytest

from metldata.builtin_transformations.infer_references.path.path_elements import (
    ReferencePathElement,
    ReferencePathElementType,
)
from metldata.builtin_transformations.infer_references.path.path_str import (
    ValidationError,
    extract_first_element,
    get_element_components,
    get_element_type,
    get_string_elements,
    get_target_class,
    path_str_to_object_elements,
    split_first_element,
    string_element_to_object,
    validate_path_str_characters,
    validate_path_str_format,
    validate_string_element,
)


@pytest.mark.parametrize(
    "path_str, is_valid",
    [
        ("class_a(class_b)>class_b", True),
        ("class_a<(class_a)class_b", True),
        ("class_1(class_2)>class_2", True),
        ("ClassA(class_b)>ClassB", True),
        ("class-a(has-class_b)>class-b", False),
        ("class_a.class_b>class_b", False),
    ],
)
def test_validate_path_str_characters(path_str: str, is_valid: bool):
    """Test the validate_path_str_characters method."""

    with nullcontext() if is_valid else pytest.raises(ValidationError):  # type: ignore
        validate_path_str_characters(path_str)


@pytest.mark.parametrize(
    "path_str, is_valid",
    [
        ("class_a(class_b)>class_b", True),
        ("class_a<(class_a)class_b", True),
        ("class_a(class_b)>class_b(class_c)>class_c", True),
        ("class_a<(class_a)class_b(class_c)>class_c", True),
        ("class_a<(class_a)class_b<(class_b)class_c", True),
        (
            "class_a(class_b)>class_b(class_c)>class_c(class_d)>class_d",
            True,
        ),
        ("class_a<(class_b)>class_b", False),
        ("class_a>class_b", False),
        ("class_a>(class_a)class_b", False),
        ("class_a(class_b)<class_b", False),
        ("class_a(class_b)>class_b(class_c)>", False),
        ("(class_b)>class_b(class_c)>class_c", False),
        ("class_a(class_b>class_b", False),
    ],
)
def test_validate_path_str_format(path_str: str, is_valid: bool):
    """Test the validate_path_str_format method."""

    with nullcontext() if is_valid else pytest.raises(ValidationError):  # type: ignore
        validate_path_str_format(path_str)


@pytest.mark.parametrize(
    "path_str, expected_first_element",
    [
        ("class_a(class_b)>class_b", "class_a(class_b)>class_b"),
        ("class_a<(class_a)class_b", "class_a<(class_a)class_b"),
        (
            "class_a(class_b)>class_b(class_c)>class_c",
            "class_a(class_b)>class_b",
        ),
    ],
)
def test_extract_first_element(path_str: str, expected_first_element: str):
    """Test the extract_first_element method."""

    observed_first_element = extract_first_element(path_str=path_str)
    assert observed_first_element == expected_first_element


@pytest.mark.parametrize(
    "path_str, expected_target_class",
    [
        ("class_a(class_b)>class_b", "class_b"),
        ("class_a<(class_a)class_b", "class_b"),
        (
            "class_a(class_b)>class_b(class_c)>class_c",
            "class_c",
        ),
    ],
)
def test_get_target_class(path_str: str, expected_target_class: str):
    """Test the get_target_class method."""

    observed_target_class = get_target_class(path_str=path_str)
    assert observed_target_class == expected_target_class


@pytest.mark.parametrize(
    "path_str, expected_first_element, expected_remaining_path",
    [
        ("class_a(class_b)>class_b", "class_a(class_b)>class_b", None),
        ("class_a<(class_a)class_b", "class_a<(class_a)class_b", None),
        (
            "class_a(class_b)>class_b(class_c)>class_c",
            "class_a(class_b)>class_b",
            "class_b(class_c)>class_c",
        ),
    ],
)
def test_split_first_element(
    path_str: str, expected_first_element: str, expected_remaining_path: Optional[str]
):
    """Test the split_first_element method."""

    observed_first_element, observed_remaining_path = split_first_element(
        path_str=path_str
    )
    assert observed_first_element == expected_first_element
    assert observed_remaining_path == expected_remaining_path


@pytest.mark.parametrize(
    "path_str, expected_elements",
    [
        ("class_a(class_b)>class_b", ["class_a(class_b)>class_b"]),
        ("class_a<(class_a)class_b", ["class_a<(class_a)class_b"]),
        (
            "class_a(class_b)>class_b(class_c)>class_c",
            [
                "class_a(class_b)>class_b",
                "class_b(class_c)>class_c",
            ],
        ),
    ],
)
def test_get_string_elements(path_str: str, expected_elements: list[str]):
    """Test the get_string_elements method."""

    observed_elements = get_string_elements(path_str=path_str)
    assert observed_elements == expected_elements


@pytest.mark.parametrize(
    "string_element, is_valid",
    [
        ("class_a(class_b)>class_b", True),
        ("class_a<(class_a)class_b", True),
        ("class_a<(class_a)>class_b", False),
        ("class_a>class_b", False),
        ("class_a(class_b)>class_b(class_c)>class_c", False),
    ],
)
def test_validate_string_element(string_element: str, is_valid: bool):
    """Test the validate_string_element method."""

    with nullcontext() if is_valid else pytest.raises(ValidationError):  # type: ignore
        validate_string_element(string_element)


@pytest.mark.parametrize(
    "string_element, expected_type",
    [
        ("class_a(class_b)>class_b", ReferencePathElementType.ACTIVE),
        ("class_a<(class_a)class_b", ReferencePathElementType.PASSIVE),
    ],
)
def test_get_element_type(string_element: str, expected_type: ReferencePathElementType):
    """Test the get_element_type method."""

    observed_type = get_element_type(string_element=string_element)
    assert observed_type == expected_type


@pytest.mark.parametrize(
    "string_element, expected_source, expected_slot, expected_target",
    [
        ("class_a(class_b)>class_b", "class_a", "class_b", "class_b"),
        ("class_a<(class_a)class_b", "class_a", "class_a", "class_b"),
    ],
)
def test_get_element_components(
    string_element: str, expected_source: str, expected_slot: str, expected_target: str
):
    """Test the get_element_components method."""

    observed_source, observed_slot, observed_target = get_element_components(
        string_element=string_element
    )
    assert observed_source == expected_source
    assert observed_slot == expected_slot
    assert observed_target == expected_target


@pytest.mark.parametrize(
    "string_element, expected_object",
    [
        (
            "class_a(class_b)>class_b",
            ReferencePathElement(
                type_=ReferencePathElementType.ACTIVE,
                source="class_a",
                slot="class_b",
                target="class_b",
            ),
        ),
        (
            "class_a<(class_a)class_b",
            ReferencePathElement(
                type_=ReferencePathElementType.PASSIVE,
                source="class_a",
                slot="class_a",
                target="class_b",
            ),
        ),
    ],
)
def test_string_element_to_object(
    string_element: str, expected_object: ReferencePathElement
):
    """Test the string_element_to_object method."""

    observed_object = string_element_to_object(string_element)
    assert observed_object == expected_object


@pytest.mark.parametrize(
    "path_str, expected_elements",
    [
        (
            "class_a(class_b)>class_b",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                )
            ],
        ),
        (
            "class_a<(class_a)class_b",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.PASSIVE,
                    source="class_a",
                    slot="class_a",
                    target="class_b",
                )
            ],
        ),
        (
            "class_a(class_b)>class_b(class_c)>class_c",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                ),
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_b",
                    slot="class_c",
                    target="class_c",
                ),
            ],
        ),
        (
            "class_a(class_b)>class_b<(class_b)class_c",
            [
                ReferencePathElement(
                    type_=ReferencePathElementType.ACTIVE,
                    source="class_a",
                    slot="class_b",
                    target="class_b",
                ),
                ReferencePathElement(
                    type_=ReferencePathElementType.PASSIVE,
                    source="class_b",
                    slot="class_b",
                    target="class_c",
                ),
            ],
        ),
    ],
)
def test_path_str_to_object_elements(
    path_str: str, expected_elements: ReferencePathElement
):
    """Test the path_str_to_object_elements method."""

    observed_elements = path_str_to_object_elements(path_str)
    assert observed_elements == expected_elements



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Basic constants and logic related to models."""

from __future__ import annotations

import dataclasses
import json
from contextlib import contextmanager
from copy import copy, deepcopy
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Generator

import jsonasobj2
import yaml
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

# The name of the root class of a model:
ROOT_CLASS = "Submission"


class MetadataModel(SchemaDefinition):
    """A dataclass for describing metadata models."""

    _schema_view = None

    @classmethod
    def init_from_path(cls, model_path: Path) -> MetadataModel:
        """Initialize from a model file in yaml format."""

        with open(model_path, "r", encoding="utf-8") as file:
            model_json = yaml.safe_load(file)

        return cls(**model_json)

    @property
    def schema_view(self) -> ExportableSchemaView:
        """Get a schema view instance from the metadata model."""
        schema_view = self._schema_view
        if schema_view is None:
            schema_view = ExportableSchemaView(self)
            self._schema_view = schema_view
        return schema_view

    def __deepcopy__(self, memo: Any):
        """Return a deep copy of the model."""
        schema_view = self._schema_view
        self._schema_view = None
        copied_model = deepcopy(super())
        self._schema_view = schema_view
        return copied_model

    def __eq__(self, other: object):
        """For comparisons."""

        if not isinstance(other, MetadataModel):
            return NotImplemented

        return self.as_dict() == other.as_dict()

    # pylint: disable=too-many-nested-blocks,too-many-branches
    def as_dict(self, essential: bool = True):  # noqa: C901
        """Get a dictionary representation of the model. If essential set to True, the
        dictionary will be cleaned of all fields that are not essential."""

        model_dict = dataclasses.asdict(self)

        if essential:
            if "classes" in model_dict:
                for class_ in model_dict["classes"].values():
                    if "from_schema" in class_:
                        del class_["from_schema"]
                    if "slot_usage" in class_:
                        slot_usage: Any = (
                            jsonasobj2.as_dict(class_["slot_usage"])
                            if isinstance(class_["slot_usage"], jsonasobj2.JsonObj)
                            else class_["slot_usage"]
                        )
                        for slot in slot_usage.values():
                            if "alias" in slot:
                                del slot["alias"]
                            if "from_schema" in slot:
                                del slot["from_schema"]
                            if "domain_of" in slot:
                                del slot["domain_of"]
            if "slots" in model_dict:
                for slot in model_dict["slots"].values():
                    if "from_schema" in slot:
                        del slot["from_schema"]
            if "enums" in model_dict:
                for enum in model_dict["enums"].values():
                    if "from_schema" in enum:
                        del enum["from_schema"]

        return model_dict

    def as_json(self) -> str:
        """Get a json representation of the model."""

        return json.dumps(self.as_dict(essential=True), indent=2)

    def as_yaml(self) -> str:
        """Get a yaml representation of the model."""

        return yaml.safe_dump(self.as_dict(essential=True))

    def write_yaml(self, path: Path) -> None:
        """Write the model to a yaml file."""

        with open(path, "w", encoding="utf-8") as file:
            yaml.safe_dump(self.as_dict(essential=True), file)

    @contextmanager
    def temporary_yaml_path(self) -> Generator[Path, None, None]:
        """Returns a context manager that creates a temporary yaml file containing the
        model and returns its path on __enter__ and deletes it on __exit__.

        This is required because some tools in the linkml schema ecosystem only support
        working with paths to yaml file and not with in-memory representations.
        """

        with NamedTemporaryFile(mode="w", encoding="utf-8") as file:
            model_json = self.as_dict()
            yaml.safe_dump(model_json, file)
            file.flush()
            yield Path(file.name)


class ExportableSchemaView(SchemaView):
    """Extend the SchemaView by adding a method for exporting a MetadataModel."""

    def __copy__(self):
        """Return a copy of the model.

        Please note, the copy will have a new uuid used for hashing.
        """

        return ExportableSchemaView(
            schema=copy(self.schema),
            importmap=copy(self.importmap),
        )

    def __deepcopy__(self, memo: Any):
        """Return a deepcopy of the model.

        Please note, the copy will have a new uuid used for hashing.
        """

        return ExportableSchemaView(
            schema=deepcopy(self.schema), importmap=copy(self.importmap)
        )

    def export_model(self) -> MetadataModel:
        """Export a MetadataModel."""

        model_json = dataclasses.asdict(deepcopy(self.schema))

        return MetadataModel(**model_json)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utilities for manipulating the schemas. This extends the functionality provided by
the standard linkml SchemaView class."""

from copy import deepcopy
from typing import Optional, Union

import jsonasobj2
from linkml_runtime.linkml_model.meta import ClassDefinition, SlotDefinition

from metldata.model_utils.anchors import AnchorPoint
from metldata.model_utils.assumptions import ROOT_CLASS
from metldata.model_utils.essentials import ExportableSchemaView
from metldata.model_utils.identifiers import get_class_identifier


class ModelManipulationError(RuntimeError):
    """A base for exception occurring in the context of model manipulations."""


class ClassNotFoundError(ModelManipulationError):
    """Raised when a class was not found."""

    def __init__(self, class_name: str):
        message = f"The class '{class_name}' does not exist in the metadata model."
        super().__init__(message)


class ClassSlotNotFoundError(ModelManipulationError):
    """Raised when a slot of a class was not found."""

    def __init__(
        self,
        class_name: str,
        slot_name: str,
        direct: bool = False,
        context: Optional[str] = None,
    ):
        message = (
            f"The slot '{slot_name}' does not exist"
            + (" as direct slot" if direct else "")
            + f" in the class '{class_name}' of the metadata model."
            + (f" {context}" if context else "")
        )
        super().__init__(message)


def get_normalized_slot_usage(*, class_: ClassDefinition) -> dict[str, SlotDefinition]:
    """Get a normalized slot usage dictionary from a class definition."""

    if class_.slot_usage:
        if isinstance(class_.slot_usage, jsonasobj2.JsonObj):
            return jsonasobj2.as_dict(class_.slot_usage)
        if isinstance(class_.slot_usage, dict):
            return deepcopy(class_.slot_usage)
        raise RuntimeError(f"Unexpected slot usage for class '{class_.name}'")

    return {}


def add_slot_if_not_exists(
    schema_view: ExportableSchemaView, new_slot: SlotDefinition
) -> ExportableSchemaView:
    """Add the new slot only if it does not exist and return a modified copy of the
    provided schema view. If it does already exist, the schema view is returned
    unmodified."""

    if not schema_view.get_slot(slot_name=new_slot.name):
        schema_view_copy = deepcopy(schema_view)
        schema_view_copy.add_slot(new_slot)

        return schema_view_copy

    return schema_view


def upsert_class_slot(
    schema_view: ExportableSchemaView, class_name: str, new_slot: SlotDefinition
) -> ExportableSchemaView:
    """Insert a new or update an existing slot of a class.

    Raises:
        ClassNotFoundError: if the specified class does not exist.
    """

    class_ = schema_view.get_class(class_name=class_name)

    if not class_:
        raise ClassNotFoundError(class_name=class_name)

    class_copy = deepcopy(class_)

    # add slot to list of slots used by the class:
    if class_copy.slots:
        if all(str(slot) != new_slot.name for slot in class_copy.slots):
            class_copy.slots.append(new_slot.name)
    else:
        class_copy.slots = [new_slot.name]

    # update slot usage:
    class_copy.slot_usage = get_normalized_slot_usage(class_=class_copy)
    class_copy.slot_usage[new_slot.name] = new_slot

    # add updated class and a global definition of the slot to a copy of schema view:
    schema_view_copy = deepcopy(schema_view)
    schema_view_copy = add_slot_if_not_exists(
        schema_view=schema_view_copy, new_slot=new_slot
    )
    schema_view_copy.add_class(class_copy)

    # workaround for a problem in linkml-runtime after converting slot usage to dict:
    schema_view_copy = schema_view_copy.export_model().schema_view

    return schema_view_copy


def delete_class_slot(
    schema_view: ExportableSchemaView, class_name: str, slot_name: str
) -> ExportableSchemaView:
    """Delete a slot from a class. This slot may not be inherited but must be
    a slot of the class itself.

    Raises:
        ClassNotFoundError: if the specified class does not exist.
        ClassSlotNotFoundError: if the specified slot does not exist in the specified class.
    """

    class_ = schema_view.get_class(class_name=class_name)

    if not class_:
        raise ClassNotFoundError(class_name=class_name)

    class_copy = deepcopy(class_)

    # modify class slots:
    all_slots = schema_view.class_slots(class_name=class_name, direct=True)

    if slot_name not in all_slots:
        raise ClassSlotNotFoundError(
            class_name=class_name,
            slot_name=slot_name,
            direct=True,
            context=(
                "Please note that you cannot delete slots that are inherited from a"
                + " parent class."
            ),
        )

    all_slots_modified = [slot for slot in all_slots if slot != slot_name]

    class_copy.slots = all_slots_modified

    # update slot usage:
    class_copy.slot_usage = get_normalized_slot_usage(class_=class_copy)
    if slot_name in class_copy.slot_usage:
        del class_copy.slot_usage[slot_name]

    # add updated class to a copy of schema view:
    schema_view_copy = deepcopy(schema_view)
    schema_view_copy.add_class(class_copy)

    return schema_view_copy


def add_slot_usage_annotation(
    *,
    schema_view: ExportableSchemaView,
    slot_name: str,
    class_name: str,
    annotation_key: str,
    annotation_value: Union[str, bool, int, float],
) -> ExportableSchemaView:
    """Add annotations to a slot in the context of a class.
    For more details see:
    https://linkml.io/linkml/schemas/metadata.html#arbitrary-annotations
    """

    class_ = schema_view.get_class(class_name=class_name)

    all_slots = schema_view.class_slots(class_name=class_name)
    if slot_name not in all_slots:
        raise ClassSlotNotFoundError(class_name=class_name, slot_name=slot_name)

    if not class_:
        raise ClassNotFoundError(class_name=class_name)

    class_copy = deepcopy(class_)

    if not class_copy.slots:
        class_copy.slots = []
    if not isinstance(class_copy.slots, list):
        raise RuntimeError(f"Unexpected slots for class '{class_name}'")
    if slot_name not in class_copy.slots:
        class_copy.slots.append(slot_name)

    class_copy.slot_usage = get_normalized_slot_usage(class_=class_copy)
    if slot_name not in class_copy.slot_usage:
        class_copy.slot_usage[slot_name] = SlotDefinition(name=slot_name)

    slot_usage = class_copy.slot_usage[slot_name]

    if slot_usage.annotations:
        slot_usage.annotations[annotation_key] = annotation_value
    else:
        slot_usage.annotations = {annotation_key: annotation_value}

    # add updated class to a copy of schema view:
    schema_view_copy = deepcopy(schema_view)
    schema_view_copy.add_class(class_copy)

    return schema_view_copy


def _get_root_class(*, schema_view: ExportableSchemaView) -> ClassDefinition:
    """A helper function to get the root class of the model."""

    root_class = schema_view.get_class(class_name=ROOT_CLASS)

    if not root_class:
        raise ClassNotFoundError(class_name=ROOT_CLASS)

    return root_class


def upsert_class(
    *, schema_view: ExportableSchemaView, class_definition: ClassDefinition
) -> ExportableSchemaView:
    """Return an updated schema view with the provided class definition being added or
    updated."""

    modified_schema_view = deepcopy(schema_view)
    modified_schema_view.schema.classes[class_definition.name] = class_definition
    modified_schema_view.set_modified()

    return modified_schema_view


def add_anchor_point(
    *,
    schema_view: ExportableSchemaView,
    anchor_point: AnchorPoint,
    description: Optional[str] = None,
) -> ExportableSchemaView:
    """Add an anchor point for a class to the tree root of the model."""

    class_ = schema_view.get_class(class_name=anchor_point.target_class)

    if not class_:
        raise ClassNotFoundError(class_name=anchor_point.target_class)

    root_slot_definition = SlotDefinition(
        name=anchor_point.root_slot,
        range=anchor_point.target_class,
        multivalued=True,
        required=True,
        inlined=True,
        inlined_as_list=True,
        description=description,
    )

    root_class = _get_root_class(schema_view=schema_view)
    return upsert_class_slot(
        schema_view=schema_view,
        class_name=root_class.name,
        new_slot=root_slot_definition,
    )


def disable_identifier_slot(
    *, schema_view: ExportableSchemaView, class_name: str
) -> ExportableSchemaView:
    """Disable the identifier for a class. This will not remove the identifier slot
    but will set its 'identifier' property to False.
    """

    identifier_slot_name = get_class_identifier(
        model=schema_view.export_model(), class_name=class_name
    )

    if not identifier_slot_name:
        return schema_view

    class_ = schema_view.get_class(class_name=class_name)

    if not class_:
        raise ClassNotFoundError(class_name=class_name)

    class_copy = deepcopy(class_)

    class_copy.slot_usage = get_normalized_slot_usage(class_=class_copy)

    if identifier_slot_name in class_copy.slot_usage:
        identifier_slot_definition = class_copy.slot_usage[identifier_slot_name]
        identifier_slot_definition.identifier = False
    else:
        identifier_slot_definition = SlotDefinition(
            name=identifier_slot_name, identifier=False
        )
        class_copy.slot_usage[identifier_slot_name] = identifier_slot_definition

    schema_view_modified = deepcopy(schema_view)
    schema_view_modified.add_class(class_copy)

    return schema_view_modified



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A self expanding defaultdict which allows to address arbitrary data paths."""

from collections import defaultdict
from typing import Any


class ExpandingDict(defaultdict):
    """Rudimentary implementation of a self expanding default dict with no error
    checking"""

    def __init__(self):
        """Create a new ExpandingDict"""
        defaultdict.__init__(self, self.__class__)

    def __resolve_path(self, path: str) -> tuple:
        """Resolves the given string representation of a path by dot-splitting
        the path. Returns the terminal key and the pre-terminal dictionary.

        Args:
            path (str): A string representation of path. Example: foo.bar.baz

        Returns:
            tuple: (dict, key) the pre-terminal dictionary and the terminal key
        """
        cur = self
        nodes = path.split(".")
        for node in nodes[:-1]:
            cur = cur[node]
        return cur, nodes[-1]

    def set_path_value(self, path: str, value: Any) -> None:
        """Set a value in the expanding dict based on a string representation of
        the key path.

        Warning: Does not perform type checking of intermediate data structures.

        Args:
            path (str): A key path. Example: foo.bar.baz
            value (Any): A value to set
        """
        holder, key = self.__resolve_path(path)
        holder[key] = value

    def get_path_value(self, path: str) -> Any:
        """Get a value in the expanding dict based on a string representation of
        the key path.

        Warning: Does not perform type checking of intermediate data structures.

        Args:
            path (str): A key path. Example: foo.bar.baz

        Returns:
            Any: The value.
        """
        holder, key = self.__resolve_path(path)
        return holder[key]

    def to_dict(self) -> dict:
        """Recursively convert the expanding dict to a regular dictionary"""
        return {
            key: value.to_dict() if isinstance(value, ExpandingDict) else value
            for key, value in self.items()
        }



#!/usr/bin/env python3

# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates a JSON schema from the service's Config class as well as a corresponding
example config yaml (or check whether these files are up to date).
"""

import importlib
import json
import subprocess
import sys
from difflib import unified_diff
from pathlib import Path
from typing import Any, Type

import yaml
from pydantic import BaseSettings
from script_utils.cli import echo_failure, echo_success, run

HERE = Path(__file__).parent.resolve()
REPO_ROOT_DIR = HERE.parent
DEV_CONFIG_YAML = REPO_ROOT_DIR / ".devcontainer" / ".dev_config.yaml"
GET_PACKAGE_NAME_SCRIPT = HERE / "get_package_name.py"
EXAMPLE_CONFIG_YAML = REPO_ROOT_DIR / "example_config.yaml"
CONFIG_SCHEMA_JSON = REPO_ROOT_DIR / "config_schema.json"


class ValidationError(RuntimeError):
    """Raised when validation of config documentation fails."""


def get_config_class() -> Type[BaseSettings]:
    """
    Dynamically imports and returns the Config class from the current service.
    This makes the script service repo agnostic.
    """
    # get the name of the microservice package
    with subprocess.Popen(
        args=[GET_PACKAGE_NAME_SCRIPT],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    ) as process:
        assert (
            process.wait() == 0 and process.stdout is not None
        ), "Failed to get package name."
        package_name = process.stdout.read().decode("utf-8").strip("\n")

    # import the Config class from the microservice package:
    config_module: Any = importlib.import_module(f"{package_name}.config")
    config_class = config_module.Config

    return config_class


def get_dev_config():
    """Get dev config object."""
    config_class = get_config_class()
    return config_class(config_yaml=DEV_CONFIG_YAML)


def get_schema() -> str:
    """Returns a JSON schema generated from a Config class."""

    config = get_dev_config()
    return config.schema_json(indent=2)


def get_example() -> str:
    """Returns an example config YAML."""

    config = get_dev_config()
    normalized_config_dict = json.loads(config.json())
    return yaml.dump(normalized_config_dict)  # pyright: ignore


def update_docs():
    """Update the example config and config schema files documenting the config
    options."""

    example = get_example()
    with open(EXAMPLE_CONFIG_YAML, "w", encoding="utf-8") as example_file:
        example_file.write(example)

    schema = get_schema()
    with open(CONFIG_SCHEMA_JSON, "w", encoding="utf-8") as schema_file:
        schema_file.write(schema)


def print_diff(expected: str, observed: str):
    """Print differences between expected and observed files."""
    echo_failure("Differences in Config YAML:")
    for line in unified_diff(
        expected.splitlines(keepends=True),
        observed.splitlines(keepends=True),
        fromfile="expected",
        tofile="observed",
    ):
        print("   ", line.rstrip())


def check_docs():
    """Check whether the example config and config schema files documenting the config
    options are up to date.

    Raises:
        ValidationError: if not up to date.
    """

    example_expected = get_example()
    with open(EXAMPLE_CONFIG_YAML, "r", encoding="utf-8") as example_file:
        example_observed = example_file.read()
    if example_expected != example_observed:
        print_diff(example_expected, example_observed)
        raise ValidationError(
            f"Example config YAML at '{EXAMPLE_CONFIG_YAML}' is not up to date."
        )

    schema_expected = get_schema()
    with open(CONFIG_SCHEMA_JSON, "r", encoding="utf-8") as schema_file:
        schema_observed = schema_file.read()
    if schema_expected != schema_observed:
        raise ValidationError(
            f"Config schema JSON at '{CONFIG_SCHEMA_JSON}' is not up to date."
        )


def main(check: bool = False):
    """Update or check the config documentation files."""

    if check:
        try:
            check_docs()
        except ValidationError as error:
            echo_failure(f"Validation failed: {error}")
            sys.exit(1)
        echo_success("Config docs are up to date.")
        return

    update_docs()
    echo_success("Successfully updated the config docs.")


if __name__ == "__main__":
    run(main)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Compose the main app and resolve dependencies."""

from fastapi import FastAPI
from ghga_service_commons.api import configure_app
from hexkit.providers.mongodb import MongoDbDaoFactory

from metldata.load.aggregator import MongoDbAggregator
from metldata.load.api import rest_api_factory
from metldata.load.config import ArtifactLoaderAPIConfig


async def get_app(config: ArtifactLoaderAPIConfig) -> FastAPI:
    """Get the FastAPI app for loading artifacts. Performs dependency injection."""

    app = FastAPI(
        title="Artifacts Loader",
        description="Load artifacts into services for querying via user-accessible API.",
    )
    configure_app(app=app, config=config)
    dao_factory = MongoDbDaoFactory(config=config)
    db_aggregator = MongoDbAggregator(config=config)

    api_router = await rest_api_factory(
        artifact_infos=config.artifact_infos,
        primary_artifact_name=config.primary_artifact_name,
        config=config,
        dao_factory=dao_factory,
        db_aggregator=db_aggregator,
        token_hashes=config.loader_token_hashes,
    )

    app.include_router(api_router)

    return app



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models to describe transformations and workflows."""

from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from graphlib import CycleError, TopologicalSorter
from typing import Callable, Generic, Optional, TypeVar

from pydantic import BaseModel, Field, create_model, root_validator, validator

from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation

# shortcuts:
# pylint: disable=unused-import
from metldata.model_utils.assumptions import MetadataModelAssumptionError  # noqa: F401
from metldata.model_utils.essentials import MetadataModel


class MetadataModelTransformationError(RuntimeError):
    """Raised when a transformation failed when applied to the metadata model."""


class MetadataTransformationError(RuntimeError):
    """Raised when a transformation failed when applied to metadata."""


Config = TypeVar("Config", bound=BaseModel)


class MetadataTransformer(ABC, Generic[Config]):
    """A base class for a metadata transformer."""

    def __init__(
        self,
        *,
        config: Config,
        original_model: MetadataModel,
        transformed_model: MetadataModel,
    ):
        """Initialize the transformer with config params, the original model, and the
        transformed model."""

        self._config = config
        self._original_model = original_model
        self._transformed_model = transformed_model

    @abstractmethod
    def transform(self, *, metadata: Json, annotation: SubmissionAnnotation) -> Json:
        """Transforms metadata.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """
        ...


@dataclass(frozen=True)
class TransformationDefinition(Generic[Config]):
    """A model for describing a transformation."""

    config_cls: type[Config] = Field(
        ..., description="The config model of the transformation."
    )
    check_model_assumptions: Callable[[MetadataModel, Config], None] = Field(
        ...,
        description=(
            "A function that checks the assumptions made about the input model."
            "Raises a MetadataModelAssumptionError if the assumptions are not met."
        ),
    )
    transform_model: Callable[[MetadataModel, Config], MetadataModel] = Field(
        ...,
        description=(
            "A function to transform the model. Raises a"
            + " MetadataModelTransformationError if the transformation fails."
        ),
    )
    metadata_transformer_factory: type[MetadataTransformer[Config]] = Field(
        ...,
        description=(
            "A class for transforming metadata. Raises a MetadataTransformationError"
            "if the transformation fails."
        ),
    )


class WorkflowConfig(BaseModel, ABC):
    """A base class for workflow configs."""


class WorkflowStepBase(BaseModel, ABC):
    """A base class for workflow steps."""

    description: str = Field(..., description="A description of the step.")
    input: Optional[str] = Field(
        ...,
        description=(
            "The name of the workflow step from which the output is used as input"
            " for this step. If this is the first step, set to None."
        ),
    )

    class Config:
        """Config for the workflow step."""

        frozen = True


class WorkflowStep(WorkflowStepBase):
    """A single step in a transformation workflow."""

    transformation_definition: TransformationDefinition = Field(
        ...,
        description="The transformation to be executed in this step.",
    )


class WorkflowDefinition(BaseModel):
    """A definition of a transformation workflow."""

    description: str = Field(..., description="A description of the workflow.")
    steps: dict[str, WorkflowStep] = Field(
        ...,
        description=(
            "A dictionary of workflow steps. The keys are the names of the steps, and"
            + " the values are the workflow steps themselves."
        ),
    )
    artifacts: dict[str, str] = Field(
        ...,
        description=(
            "A dictionary of artifacts that are output by this workflow."
            + " The keys are the names of the artifacts, and the values are the names"
            + " of the workflow steps that output them."
        ),
    )

    # pylint: disable=no-self-argument
    @validator("steps", pre=False)
    def validate_step_references(
        cls, steps: dict[str, WorkflowStep]
    ) -> dict[str, WorkflowStep]:
        """Validate that workflow steps reference other existing steps as input.
        There should be exactly one step with input=None.
        """

        step_with_no_input_found = False

        for step_name, step in steps.items():
            if step.input is None:
                if step_with_no_input_found:
                    raise ValueError(
                        "There should be exactly one step with input=None. But multiple"
                        + " were found."
                    )
                step_with_no_input_found = True
                continue
            if step.input not in steps:
                raise ValueError(
                    f"Step {step.input} referenced in step {step_name} is not defined."
                )

        if not step_with_no_input_found:
            raise ValueError(
                "There should be exactly one step with input=None but none was found."
            )

        return steps

    @root_validator(pre=False)
    def validate_artifact_references(cls, values):
        """Validate that artifacts reference existing workflow steps."""

        steps = values.get("steps")
        if steps is None:
            raise ValueError("Steps are undefined.")
        artifacts = values.get("artifacts")
        if artifacts is None:
            raise ValueError("Artifacts are undefined.")

        for artifact_name, step_name in artifacts.items():
            if step_name not in steps:
                raise ValueError(
                    f"Step {step_name} referenced in artifact {artifact_name} is not defined."
                )

        return values

    @property
    def config_cls(self) -> type[WorkflowConfig]:
        """Get a config model containing the config requirements from all workflow
        steps."""

        step_configs = {
            step_name: (step.transformation_definition.config_cls, ...)
            for step_name, step in self.steps.items()
        }

        config_cls = create_model(  # type: ignore
            "SpecificWorkflowConfig",
            **step_configs,
            __base__=WorkflowConfig,
        )

        return config_cls

    @property
    def step_order(self) -> list[str]:
        """Get a list of step names in the order in which the steps should be executed."""

        # create graph from steps
        graph: dict[str, set[str]] = defaultdict(set[str])
        for step_name, step in self.steps.items():
            if step.input:
                graph[step_name].add(step.input)

        # sort with TopologicalSorter
        topological_sorter = TopologicalSorter(graph)
        try:
            return list(topological_sorter.static_order())
        except CycleError as exc:
            raise RuntimeError("Step definitions imply a circular dependency.") from exc

    class Config:
        """Config for the workflow step."""

        frozen = True



#!/usr/bin/env python3

# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate documentation for this package using different sources."""

import json
import subprocess  # nosec
import sys
from pathlib import Path
from string import Template

import jsonschema2md
from pydantic import BaseModel, Field
from script_utils.cli import echo_failure, echo_success, run
from setuptools.config.setupcfg import read_configuration
from stringcase import spinalcase, titlecase

ROOT_DIR = Path(__file__).parent.parent.resolve()
SETUP_CFG_PATH = ROOT_DIR / "setup.cfg"
DESCRIPTION_PATH = ROOT_DIR / ".description.md"
DESIGN_PATH = ROOT_DIR / ".design.md"
README_TEMPLATE_PATH = ROOT_DIR / ".readme_template.md"
CONFIG_SCHEMA_PATH = ROOT_DIR / "config_schema.json"
OPENAPI_YAML_REL_PATH = "./openapi.yaml"
README_PATH = ROOT_DIR / "README.md"


class PackageHeader(BaseModel):
    """A basic summary of a package."""

    shortname: str = Field(
        ...,
        description=(
            "The abbreviation of the package name. Is identical to the package name."
        ),
    )
    version: str = Field(..., description="The version of the package.")
    summary: str = Field(
        ..., description="A short 1 or 2 sentence summary of the package."
    )


class PackageName(BaseModel):
    """The name of a package and it's different representations."""

    name: str = Field(..., description="The full name of the package in spinal case.")
    title: str = Field(..., description="The name of the package formatted as title.")


class PackageDetails(PackageHeader, PackageName):
    """A container for details on a package used to build documentation."""

    description: str = Field(
        ..., description="A markdown-formatted description of the package."
    )
    design_description: str = Field(
        ...,
        description=(
            "A markdown-formatted description of overall architecture and design of"
            + " the package."
        ),
    )
    config_description: str = Field(
        ...,
        description=(
            "A markdown-formatted list of all configuration parameters of this package."
        ),
    )
    openapi_doc: str = Field(
        ...,
        description=(
            "A markdown-formatted description rendering or linking to an OpenAPI"
            " specification of the package."
        ),
    )


def read_package_header() -> PackageHeader:
    """Read basic information about the package from the setup.cfg."""

    setup_config = read_configuration(SETUP_CFG_PATH)
    setup_metadata = setup_config["metadata"]
    return PackageHeader(
        shortname=setup_metadata["name"],
        version=setup_metadata["version"],
        summary=setup_metadata["description"],
    )


def read_package_name() -> PackageName:
    """Infer the package name from the name of the git origin."""

    with subprocess.Popen(
        args="basename -s .git `git config --get remote.origin.url`",
        cwd=ROOT_DIR,
        stdout=subprocess.PIPE,
        shell=True,
    ) as process:
        stdout, _ = process.communicate()

    if not stdout:
        raise RuntimeError("The name of the git origin could not be resolved.")
    git_origin_name = stdout.decode("utf-8").strip()

    return PackageName(
        name=spinalcase(git_origin_name), title=titlecase(git_origin_name)
    )


def read_package_description() -> str:
    """Read the package description."""

    return DESCRIPTION_PATH.read_text()


def read_design_description() -> str:
    """Read the design description."""

    return DESIGN_PATH.read_text()


def generate_config_docs() -> str:
    """Generate markdown-formatted documentation for the configration parameters
    listed in the config schema."""

    parser = jsonschema2md.Parser(
        examples_as_yaml=False,
        show_examples="all",
    )
    with open(CONFIG_SCHEMA_PATH, "r", encoding="utf-8") as json_file:
        config_schema = json.load(json_file)

    md_lines = parser.parse_schema(config_schema)

    # ignore everything before the properites header:
    properties_index = md_lines.index("## Properties\n\n")
    md_lines = md_lines[properties_index + 1 :]

    return "\n".join(md_lines)


def generate_openapi_docs() -> str:
    """Generate markdown-formatted documentation linking to or rendering an OpenAPI
    specification of the package. If no OpenAPI specification is present, return an
    empty string."""

    open_api_yaml_path = ROOT_DIR / OPENAPI_YAML_REL_PATH

    if not open_api_yaml_path.exists():
        return ""

    return (
        "## HTTP API\n"
        + "An OpenAPI specification for this service can be found"
        + f" [here]({OPENAPI_YAML_REL_PATH})."
    )


def get_package_details() -> PackageDetails:
    """Get details required to build documentation for the package."""

    header = read_package_header()
    name = read_package_name()
    description = read_package_description()
    config_description = generate_config_docs()
    return PackageDetails(
        **header.dict(),
        **name.dict(),
        description=description,
        config_description=config_description,
        design_description=read_design_description(),
        openapi_doc=generate_openapi_docs(),
    )


def generate_single_readme(*, details: PackageDetails) -> str:
    """Generate a single markdown-formatted readme file for the package based on the
    provided details."""

    template_content = README_TEMPLATE_PATH.read_text()
    template = Template(template_content)
    return template.substitute(details.dict())


def main(check: bool = False) -> None:
    """Update the readme markdown."""

    details = get_package_details()
    readme_content = generate_single_readme(details=details)

    if check:
        if README_PATH.read_text() != readme_content:
            echo_failure("README.md is not up to date.")
            sys.exit(1)
        echo_success("README.md is up to date.")
        return

    README_PATH.write_text(readme_content)
    echo_success("Successfully updated README.md.")


if __name__ == "__main__":
    run(main)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Config parameters and their defaults."""

from pydantic import BaseSettings, Field, validator

from metldata.artifacts_rest.models import ArtifactInfo


class ArtifactsRestConfig(BaseSettings):
    """Config parameters and their defaults."""

    artifact_infos: list[ArtifactInfo] = Field(
        ...,
        description="Information for artifacts to be queryable via the Artifacts REST API.",
    )

    # pylint: disable=no-self-argument
    @validator("artifact_infos")
    def validate_artifact_info_names(
        cls, value: list[ArtifactInfo]
    ) -> list[ArtifactInfo]:
        """Validate that artifact names are unique."""

        artifact_names = [artifact_info.name for artifact_info in value]
        if len(artifact_names) != len(set(artifact_names)):
            raise ValueError("Artifact names must be unique.")

        return value



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for loading artifacts."""

from typing import cast

from ghga_event_schemas.pydantic_ import (
    MetadataDatasetFile,
    MetadataDatasetOverview,
    MetadataDatasetStage,
    SearchableResource,
)
from ghga_service_commons.utils.files import get_file_extension

from metldata.artifacts_rest.artifact_dao import ArtifactDaoCollection
from metldata.artifacts_rest.models import (
    ArtifactInfo,
    ArtifactResource,
    ArtifactResourceClass,
)
from metldata.custom_types import Json
from metldata.load.event_publisher import EventPublisherPort
from metldata.metadata_utils import (
    SlotNotFoundError,
    get_resources_of_class,
    lookup_self_id,
    lookup_slot_in_resource,
)


def extract_class_resources_from_artifact(
    *, artifact_content: Json, resource_class: ArtifactResourceClass
) -> list[ArtifactResource]:
    """Extract the resources from the given artifact content for the given class."""

    resource_jsons = get_resources_of_class(
        class_name=resource_class.name,
        global_metadata=artifact_content,
        anchor_points_by_target={resource_class.name: resource_class.anchor_point},
    )

    return [
        ArtifactResource(
            id_=lookup_self_id(
                resource=resource_json,
                identifier_slot=resource_class.anchor_point.identifier_slot,
            ),
            class_name=resource_class.name,
            content=resource_json,
        )
        for resource_json in resource_jsons
    ]


def extract_all_resources_from_artifact(
    artifact_content: Json, artifact_info: ArtifactInfo
) -> list[ArtifactResource]:
    """Extract all resources from the given artifact content for all resource classes
    as specified in the artifact info."""

    return [
        resource
        for resource_class in artifact_info.resource_classes.values()
        for resource in extract_class_resources_from_artifact(
            artifact_content=artifact_content,
            resource_class=resource_class,
        )
    ]


async def save_artifact_resource(
    *,
    resource: ArtifactResource,
    artifact_name: str,
    dao_collection: ArtifactDaoCollection,
) -> None:
    """Save the given resource into the database using the given DAO collection."""

    dao = await dao_collection.get_dao(
        artifact_name=artifact_name, class_name=resource.class_name
    )
    await dao.upsert(resource)


async def remove_artifact_resource(
    *,
    resource_id: str,
    class_name: str,
    artifact_name: str,
    dao_collection: ArtifactDaoCollection,
) -> None:
    """Remove a given artifact resource from the database using the given DAO collection"""

    dao = await dao_collection.get_dao(
        artifact_name=artifact_name, class_name=class_name
    )
    await dao.delete(id_=resource_id)


async def load_artifact_resources(
    *,
    artifact_content: Json,
    artifact_info: ArtifactInfo,
    dao_collection: ArtifactDaoCollection,
) -> None:
    """Load the resources from the given artifacts into the database using the given
    DAO collection.
    """

    resources = extract_all_resources_from_artifact(
        artifact_content=artifact_content, artifact_info=artifact_info
    )

    for resource in resources:
        await save_artifact_resource(
            resource=resource,
            artifact_name=artifact_info.name,
            dao_collection=dao_collection,
        )


async def process_removed_resources(
    *,
    event_publisher: EventPublisherPort,
    resource_tags: set[tuple[str, str, str]],
    dao_collection: ArtifactDaoCollection,
):
    """Delete no longer needed artifact resources from DB and send corresponding events"""

    for resource_tag in resource_tags:
        artifact_name, class_name, resource_id = resource_tag
        # resource tag was obtained from querying the db, so resource with given ID
        # should be present
        await remove_artifact_resource(
            artifact_name=artifact_name,
            class_name=class_name,
            resource_id=resource_id,
            dao_collection=dao_collection,
        )

        await event_publisher.process_resource_deletion(
            accession=resource_id, class_name=class_name
        )
        if event_publisher.is_primary_dataset_source(
            artifact_name=artifact_name, resource_class_name=class_name
        ):
            await event_publisher.process_dataset_deletion(accession=resource_id)


async def process_new_or_changed_resources(
    *,
    artifact_info_dict: dict[str, ArtifactInfo],
    event_publisher: EventPublisherPort,
    resources: dict[tuple[str, str, str], ArtifactResource],
    dao_collection: ArtifactDaoCollection,
):
    """Insert newly received artifact resources into DB and send corresponding events"""

    for resource_tag, resource in resources.items():
        artifact_name = resource_tag[0]
        # no resource tag was obtained from querying the db, so the resource with the
        # given ID should not be present
        await save_artifact_resource(
            resource=resource,
            artifact_name=artifact_name,
            dao_collection=dao_collection,
        )

        if event_publisher.is_primary_dataset_source(
            artifact_name=artifact_name, resource_class_name=resource.class_name
        ):
            await process_resource_upsert(
                artifact_info_dict=artifact_info_dict,
                artifact_name=artifact_name,
                event_publisher=event_publisher,
                resource=resource,
            )


async def process_resource_upsert(  # pylint: disable=too-many-locals
    *,
    artifact_info_dict: dict[str, ArtifactInfo],
    artifact_name: str,
    event_publisher: EventPublisherPort,
    resource: ArtifactResource,
):
    """Convert available data to correct event model and delegate firing appropriate events"""

    artifact_info = artifact_info_dict[artifact_name]
    file_slots = get_file_slots(artifact_info=artifact_info, resource=resource)

    metadata_dataset_files = convert_file_information(file_slots=file_slots)

    dataset_description = cast(
        str,
        lookup_slot_in_resource(resource=resource.content, slot_name="description"),
    )
    dataset_title = cast(
        str,
        lookup_slot_in_resource(resource=resource.content, slot_name="title"),
    )

    dataset_overview = MetadataDatasetOverview(
        accession=resource.id_,
        title=dataset_title,
        stage=MetadataDatasetStage.DOWNLOAD,
        description=dataset_description,
        files=metadata_dataset_files,
    )
    await event_publisher.process_dataset_upsert(dataset_overview=dataset_overview)

    searchable_resource = SearchableResource(
        accession=resource.id_, class_name=resource.class_name, content=resource.content
    )
    await event_publisher.process_resource_upsert(resource=searchable_resource)


def get_file_slots(artifact_info: ArtifactInfo, resource: ArtifactResource):
    """Get files resource slots from file class names"""

    file_slots = []
    for class_name, resource_class in artifact_info.resource_classes.items():
        if class_name.endswith("File"):
            anchor_point = resource_class.anchor_point
            slot_name = anchor_point.root_slot

            try:
                file_slot = lookup_slot_in_resource(
                    resource=resource.content, slot_name=slot_name
                )
                # file slots should be lists
                file_slot = cast(list[Json], file_slot)
            except SlotNotFoundError:
                continue

            file_slots.append(file_slot)

    return file_slots


def convert_file_information(file_slots: list[list[Json]]) -> list[MetadataDatasetFile]:
    """Convert file slot information into MetadataDatasetFiles"""
    files = []
    for file_slot in file_slots:
        for file in file_slot:
            extension = get_file_extension(filename=file["name"])

            files.append(
                MetadataDatasetFile(
                    accession=file["accession"],
                    description=None,
                    file_extension=extension,
                )
            )

    return files



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""All classes that shall be referenced by ID must be linked in the root class of the
model. The slot in the root that links to a specific class is called the anchor point of
that class. This module provides logic for handling these anchor points.
"""

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SlotDefinition
from pydantic import BaseModel, Field

from metldata.model_utils.essentials import ROOT_CLASS, MetadataModel
from metldata.model_utils.identifiers import get_class_identifiers


class InvalidAnchorPointError(RuntimeError):
    """Raised when an anchor point defined in a model is invalid."""


class AnchorPointNotFoundError(RuntimeError):
    """Raised when no anchor point was found for a specific class."""


class ClassNotAnchoredError(RuntimeError):
    """Raised when a class is not anchored."""


class AnchorPoint(BaseModel):
    """A model for describing an anchor point for the specified target class."""

    target_class: str = Field(..., description="The name of the class to be targeted.")
    identifier_slot: str = Field(
        ...,
        description=(
            "The name of the slot in the target class that is used as identifier."
        ),
    )
    root_slot: str = Field(
        ...,
        description=(
            "The name of the slot in the root class used to link to the target class."
        ),
    )

    class Config:
        """Pydantic Configs."""

        frozen = True


def check_root_slot(slot: SlotDefinition):
    """Make sure that the given root slot is a valid anchor point. Validates that the
    slot is multivalued, required, and inlined but not inlined as list.

    Raises:
        InvalidAnchorPointError: if validation fails.
    """

    if not slot.multivalued:
        raise InvalidAnchorPointError(
            f"The anchor point at the slot '{slot.name}' is not multivalued."
        )

    if slot.inlined is None:
        raise InvalidAnchorPointError(
            f"The inlined attribute for slot '{slot.name}' is not defined."
        )

    if not slot.inlined:
        raise InvalidAnchorPointError(
            f"The inlined attribute for slot '{slot.name}' is set to False,"
            + " however, slots in the root class must be inlined."
        )

    if not slot.inlined_as_list:
        raise InvalidAnchorPointError(
            f"The inlined_as_list attribute for slot '{slot.name}' is set to False,"
            + " however, slots in the root class must be inlined as list."
        )

    if not slot.required:
        raise InvalidAnchorPointError(
            f"The anchor point at the slot '{slot.name}' is not required."
        )


def get_slot_target_class(*, slot: SlotDefinition, schema_view: SchemaView) -> str:
    """Checks that the specified slot refers to another class and returns the name
    of that class."""

    if slot.range is None:
        raise InvalidAnchorPointError(
            f"The slot '{slot.name}', which is used as anchor point, has no range"
            + " defined."
        )

    target_class = schema_view.get_class(class_name=slot.range)

    if target_class is None:
        raise InvalidAnchorPointError(
            f"The range of slot '{slot.name}', which is used as anchor point, does not"
            + " point to a valid class."
        )

    return slot.range


def get_anchor_points(*, model: MetadataModel) -> set[AnchorPoint]:
    """Get all anchor points of the specified model."""

    identifiers_by_class = get_class_identifiers(model=model)

    schema_view = model.schema_view
    root_slots = schema_view.class_induced_slots(class_name=ROOT_CLASS)

    # validation root slots:
    for root_slot in root_slots:
        check_root_slot(root_slot)

    anchor_point: set[AnchorPoint] = set()
    for root_slot in root_slots:
        target_class = get_slot_target_class(slot=root_slot, schema_view=schema_view)
        identifier = identifiers_by_class[target_class]
        if not identifier:
            raise InvalidAnchorPointError(
                f"The class '{target_class}' has no identifier defined."
            )

        anchor_point.add(
            AnchorPoint(
                target_class=target_class,
                identifier_slot=identifier,
                root_slot=str(root_slot.name),
            )
        )

    return anchor_point


def get_anchors_points_by_target(*, model: MetadataModel) -> dict[str, AnchorPoint]:
    """Get a dictionary with the keys corresponding to class names and the values
    corresponding to anchor points."""

    anchor_points = get_anchor_points(model=model)

    return {anchor_point.target_class: anchor_point for anchor_point in anchor_points}


def filter_anchor_points(
    *, anchor_points_by_target: dict[str, AnchorPoint], classes_of_interest: set[str]
) -> dict[str, AnchorPoint]:
    """Filter the provided anchor points by a list of classes of interest.

    Raises:
        AnchorPointNotFoundError: if no anchor point was found for a class of interest.
    """

    # check if anchor points exists for all classes:
    classes_without_anchor_points = classes_of_interest.difference(
        anchor_points_by_target.keys()
    )
    if classes_without_anchor_points:
        raise AnchorPointNotFoundError(
            "Following classes have no anchor points: "
            + ", ".join(sorted(classes_without_anchor_points))
        )

    return {
        class_name: anchor_point
        for class_name, anchor_point in anchor_points_by_target.items()
        if class_name in classes_of_interest
    }


def lookup_anchor_point(
    *, class_name: str, anchor_points_by_target: dict[str, AnchorPoint]
) -> AnchorPoint:
    """Lookup the anchor point for the given class."""

    anchor_point = anchor_points_by_target.get(class_name)

    if anchor_point is None:
        raise AnchorPointNotFoundError(
            f"Cannot find anchor point class '{class_name}'."
        )

    return anchor_point


def invert_anchor_points_by_target(
    anchor_points_by_target: dict[str, AnchorPoint]
) -> dict[str, str]:
    """Convert the anchor points by target dict into an class by anchor point dict."""

    return {
        anchor_point.root_slot: class_name
        for class_name, anchor_point in anchor_points_by_target.items()
    }


def get_target_by_anchor_point(*, model: MetadataModel) -> dict[str, str]:
    """Get a mapping from the root slot where a class is anchored to the corresponding
    class."""

    anchor_points = get_anchor_points(model=model)

    return {
        anchor_point.root_slot: anchor_point.target_class
        for anchor_point in anchor_points
    }


def lookup_class_by_anchor_point(
    *, root_slot: str, target_by_anchor_point: dict[str, str]
) -> str:
    """Lookup the class for the given anchor point."""

    class_name = target_by_anchor_point.get(root_slot)

    if class_name is None:
        raise ClassNotAnchoredError(
            f"Cannot find class for anchor point with root slot '{root_slot}'."
        )

    return class_name



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Core functionality for aggregate transformations."""

from metldata.builtin_transformations.aggregate.cached_model import CachedMetadataModel
from metldata.builtin_transformations.aggregate.config import AggregateConfig
from metldata.builtin_transformations.aggregate.metadata_transform import (
    execute_aggregations,
)
from metldata.builtin_transformations.aggregate.model_transform import (
    build_aggregation_model,
)
from metldata.custom_types import Json
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.assumptions import check_basic_model_assumption
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import MetadataTransformer, TransformationDefinition


class AggregateTransformer(MetadataTransformer[AggregateConfig]):
    """Transformer to create summary statistics from metadata."""

    def __init__(
        self,
        config: AggregateConfig,
        original_model: MetadataModel,
        transformed_model: MetadataModel,
    ):
        """Initialize the transformer."""

        super().__init__(
            config=config,
            original_model=original_model,
            transformed_model=transformed_model,
        )

        self._original_cached_model = CachedMetadataModel(model=self._original_model)
        self._transformed_cached_model = CachedMetadataModel(
            model=self._transformed_model
        )

    def transform(self, *, metadata: Json, annotation: SubmissionAnnotation) -> Json:
        """Transforms metadata.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """
        return execute_aggregations(
            original_model=self._original_cached_model,
            transformed_anchors_points=self._transformed_cached_model.anchors_points_by_target,
            metadata=metadata,
            aggregations=self._config.aggregations,
        )


def check_model_assumptions(
    model: MetadataModel,
    config: AggregateConfig,  # pylint: disable=unused-argument
) -> None:
    """Check the assumptions of the model.

    Raises:
        MetadataModelAssumptionError:
            if the model does not fulfill the assumptions.
    """
    check_basic_model_assumption(model=model)


def transform_model(model: MetadataModel, config: AggregateConfig) -> MetadataModel:
    """Transform the metadata model.

    Raises:
        MetadataModelTransformationError:
            if the transformation fails.
    """
    return build_aggregation_model(model=model, config=config)


AGGREGATE_TRANSFORMATION = TransformationDefinition[AggregateConfig](
    config_cls=AggregateConfig,
    check_model_assumptions=check_model_assumptions,
    transform_model=transform_model,
    metadata_transformer_factory=AggregateTransformer,
)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A transformation to infer references based on existing ones in the metadata model."""

from metldata.builtin_transformations.infer_references.config import (
    ReferenceInferenceConfig,
)
from metldata.builtin_transformations.infer_references.metadata_transform import (
    add_references_to_metadata,
)
from metldata.builtin_transformations.infer_references.model_transform import (
    add_references_to_model,
)
from metldata.event_handling.models import SubmissionAnnotation
from metldata.model_utils.anchors import get_anchors_points_by_target
from metldata.model_utils.assumptions import check_basic_model_assumption
from metldata.model_utils.essentials import MetadataModel
from metldata.transform.base import Json, MetadataTransformer, TransformationDefinition


class ReferenceInferenceMetadataTransformer(
    MetadataTransformer[ReferenceInferenceConfig]
):
    """A transformer that infers references in metadata based on existing ones."""

    def __init__(
        self,
        config: ReferenceInferenceConfig,
        original_model: MetadataModel,
        transformed_model: MetadataModel,
    ):
        """Initialize the transformer."""

        super().__init__(
            config=config,
            original_model=original_model,
            transformed_model=transformed_model,
        )

        self._anchor_points_by_target = get_anchors_points_by_target(
            model=self._original_model
        )

    def transform(self, *, metadata: Json, annotation: SubmissionAnnotation) -> Json:
        """Transforms metadata.

        Args:
            metadata: The metadata to be transformed.
            annotation: The annotation on the metadata.

        Raises:
            MetadataTransformationError:
                if the transformation fails.
        """

        return add_references_to_metadata(
            metadata=metadata,
            references=self._config.inferred_references,
            anchor_points_by_target=self._anchor_points_by_target,
        )


def check_model_assumptions(
    model: MetadataModel,
    config: ReferenceInferenceConfig,  # pylint: disable=unused-argument
) -> None:
    """Check the assumptions of the model.

    Raises:
        MetadataModelAssumptionError:
            if the model does not fulfill the assumptions.
    """

    check_basic_model_assumption(model=model)


def transform_model(
    model: MetadataModel, config: ReferenceInferenceConfig
) -> MetadataModel:
    """Transform the metadata model.

    Raises:
        MetadataModelTransformationError:
            if the transformation fails.
    """

    return add_references_to_model(
        model=model,
        references=config.inferred_references,
    )


REFERENCE_INFERENCE_TRANSFORMATION = TransformationDefinition[ReferenceInferenceConfig](
    config_cls=ReferenceInferenceConfig,
    check_model_assumptions=check_model_assumptions,
    transform_model=transform_model,
    metadata_transformer_factory=ReferenceInferenceMetadataTransformer,
)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test the identifiers module."""

from metldata.model_utils.anchors import AnchorPoint
from metldata.submission_registry.identifiers import generate_accession_map


class FakeAccessionRegistry:
    """A fake accession registry."""

    def __init__(self):
        """Initialize with counter to generate predictable accessions."""

        self._counter = 1

    def get_accession(self, *, resource_type: str) -> str:
        """
        Generates and registers a new accession for a resource of the specified type.
        """

        accession = f"generated_{resource_type.lower()}_accession{self._counter}"
        self._counter += 1

        return accession


def test_generate_accession_map():
    """Test generating an accession map for a given content."""

    accession_registry = FakeAccessionRegistry()
    content = {
        "class1_anchor": [{"alias": "test_alias1"}],
        "class2_anchor": [{"alias": "test_alias2"}, {"alias": "test_alias3"}],
    }
    anchor_points_by_target = {
        "Class1": AnchorPoint(
            root_slot="class1_anchor", target_class="Class1", identifier_slot="alias"
        ),
        "Class2": AnchorPoint(
            root_slot="class2_anchor", target_class="Class2", identifier_slot="alias"
        ),
        "Class3": AnchorPoint(
            root_slot="class3_anchor", target_class="Class3", identifier_slot="alias"
        ),
    }
    existing_accession_map = {
        "class2_anchor": {"test_alias2": "existing_class2_accession1"},
        "class3_anchor": {"test_alias3": "existing_class3_accession2"},
    }
    expected_accession_map = {
        "class1_anchor": {"test_alias1": "generated_class1_accession1"},
        "class2_anchor": {
            "test_alias2": "existing_class2_accession1",
            "test_alias3": "generated_class2_accession2",
        },
    }

    observed_accession_map = generate_accession_map(
        content=content,
        existing_accession_map=existing_accession_map,
        accession_registry=accession_registry,  # type: ignore
        anchor_points_by_target=anchor_points_by_target,
    )

    assert observed_accession_map == expected_accession_map



#!/usr/bin/env python3

# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Extracts the package name from the setup.cfg"""

from pathlib import Path

REPO_ROOT_DIR = Path(__file__).parent.parent.resolve()
SETUP_CFG_PATH = REPO_ROOT_DIR / "setup.cfg"
NAME_PREFIX = "name = "


def get_package_name() -> str:
    """Extracts the package name"""

    with open(SETUP_CFG_PATH, "r", encoding="utf8") as setup_cfg:
        for line in setup_cfg.readlines():
            line_stripped = line.strip()
            if line_stripped.startswith(NAME_PREFIX):
                package_name = line_stripped[len(NAME_PREFIX) :]
                return package_name
        raise RuntimeError("Could not find package name.")


def run():
    """Run this script."""
    package_name = get_package_name()
    print(package_name)


if __name__ == "__main__":
    run()
